import warnings
from typing import Optional

import numpy as np
import numpy.typing as npt
from scipy import optimize
from scipy.optimize import OptimizeWarning

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import Bool, Switch, UnsignedFloat, UnsignedInteger
from CADETProcess.optimization import OptimizationProblem, OptimizerBase


class SciPyInterface(OptimizerBase):
    """
    Wrapper around scipy's optimization suite.

    Defines the bounds and all constraints, saved in a constraint_object. Also
    the jacobian matrix is defined for several solvers.

    Parameters
    ----------
    finite_diff_rel_step : None or array_like, optional
        Relative step size to use for the numerical approximation of the jacobian.
        The absolute step size `h` is computed as `h = rel_step * sign(x) * max(1, abs(x))`,
        possibly adjusted to fit into the bounds. For `method='3-point'`,
        the sign of `h` is ignored.
        If `None` (default), the step size is selected automatically.
    tol : float, optional
        Tolerance for termination. When tol is specified, the selected minimization
        algorithm sets some relevant solver-specific tolerance(s) equal to tol.
        For detailed control, use solver-specific options.
    jac : {'2-point', '3-point', 'cs'}
        Method for computing the gradient vector. Only applicable to specific
        solvers (CG, BFGS, Newton-CG, L-BFGS-B, TNC, SLSQP, dogleg, trust-ncg,
        trust-krylov, trust-exact, and trust-constr).
        The default is 2-point.

    See Also
    --------
    COBYLA
    TrustConstr
    NelderMead
    SLSQP
    CADETProcess.optimization.OptimizationProblem.evaluate_objectives
    options
    scipy.optimize.minimize

    """

    finite_diff_rel_step = UnsignedFloat()
    tol = UnsignedFloat()
    jac = Switch(valid=["2-point", "3-point", "cs"], default="2-point")

    def _run(
        self,
        optimization_problem: OptimizationProblem,
        x0: Optional[list] = None,
    ) -> None:
        """
        Solve the optimization problem using any of the scipy methods.

        Parameters
        ----------
        results : OptimizationResults
            Optimization results including optimization_problem and solver
            configuration.
        x0 : list, optional
            Initial values of independent variables in untransformed space.

        See Also
        --------
        COBYLA
        TrustConstr
        NelderMead
        SLSQP
        CADETProcess.optimization.OptimizationProblem.evaluate_objectives
        options
        scipy.optimize.minimize
        """
        self.n_evals = 0

        if optimization_problem.n_objectives > 1:
            raise CADETProcessError("Can only handle single objective.")

        def objective_function(x: npt.ArrayLike) -> np.ndarray:
            return optimization_problem.evaluate_objectives(
                x,
                untransform=True,
                get_dependent_values=True,
                ensure_minimization=True,
            )[0]

        def callback_function(x: npt.ArrayLike, state: dict = None) -> bool:
            """
            Report progress after evaluation.

            Notes
            -----
            Currently, this evaluates all functions again. This should not be a problem
            since objectives and constraints are automatically cached.

            Unfortunately, only `trust-constr` returns a `state` which contains the
            current best point. Hence, the internal pareto front is used.
            """
            self.n_evals += 1

            x = x.tolist()
            f = optimization_problem.evaluate_objectives(
                x,
                untransform=True,
                get_dependent_values=True,
                ensure_minimization=True,
            )
            g = optimization_problem.evaluate_nonlinear_constraints(
                x,
                untransform=True,
                get_dependent_values=True,
            )
            cv = optimization_problem.evaluate_nonlinear_constraints_violation(
                x,
                untransform=True,
                get_dependent_values=True,
            )

            self.run_post_processing(x, f, g, cv, self.n_evals)

            return False

        if x0 is None:
            x0 = optimization_problem.create_initial_values(
                1, include_dependent_variables=False
            )[0]

        x0_transformed = optimization_problem.transform(x0)

        options = self.specific_options
        if self.results.n_gen > 0:
            x0 = self.results.population_last.x[0, :]
            self.n_evals = self.results.n_evals
            options["maxiter"] = self.maxiter - self.n_evals
            if str(self) == "COBYLA":
                options["maxiter"] -= 1

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=OptimizeWarning)
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            scipy_results = optimize.minimize(
                objective_function,
                x0=x0_transformed,
                method=str(self),
                tol=self.tol,
                jac=self.jac,
                constraints=self.get_constraint_objects(optimization_problem),
                bounds=self.get_bounds(optimization_problem),
                options=options,
                callback=callback_function,
            )

        self.results.success = bool(scipy_results.success)
        self.results.exit_flag = scipy_results.status
        self.results.exit_message = scipy_results.message

    def get_bounds(self, optimization_problem: OptimizationProblem) -> optimize.Bounds:
        """
        Return the optimized bounds of a given optimization_problem as a Bound object.

        Optimizes the bounds of the optimization_problem by calling the method
        optimize.Bounds. Keep_feasible is set to True.

        Returns
        -------
        bounds : Bounds
            Returns the optimized bounds as an object called bounds.
        """
        return optimize.Bounds(
            optimization_problem.lower_bounds_independent_transformed,
            optimization_problem.upper_bounds_independent_transformed,
            keep_feasible=True,
        )

    def get_constraint_objects(self, optimization_problem: OptimizationProblem) -> list:
        """
        Return constraints as objets.

        Returns
        -------
        constraint_objects : list
            List containing lists of all constraint types of the optimization_problem.
            If type of constraints is not defined, it is replaced with None.

        See Also
        --------
        lincon_obj
        lincon_obj
        nonlincon_obj
        """
        lincon = self.get_lincon_obj(optimization_problem)
        lineqcon = self.get_lineqcon_obj(optimization_problem)
        nonlincon = self.get_nonlincon_obj(optimization_problem)

        constraints = [lincon, lineqcon, *nonlincon]

        return [con for con in constraints if con is not None]

    def get_lincon_obj(
        self, optimization_problem: OptimizationProblem
    ) -> optimize.LinearConstraint:
        """
        Return the linear constraints as an object.

        Returns
        -------
        lincon_obj : LinearConstraint
            Linear Constraint object with lower and upper bounds of b of the
            optimization_problem.

        See Also
        --------
        constraint_objects
        A
        b
        """
        if optimization_problem.n_linear_constraints == 0:
            return None

        lb = [-np.inf] * len(optimization_problem.b)
        ub = optimization_problem.b_transformed

        return optimize.LinearConstraint(
            optimization_problem.A_independent_transformed, lb, ub, keep_feasible=True
        )

    def get_lineqcon_obj(
        self, optimization_problem: OptimizationProblem
    ) -> optimize.LinearConstraint:
        """
        Return the linear equality constraints as an object.

        Returns
        -------
        lineqcon_obj : LinearConstraint
            Linear equality Constraint object with lower and upper bounds of beq of the
            optimization_problem.

        See Also
        --------
        constraint_objects
        Aeq
        beq
        """
        if optimization_problem.n_linear_equality_constraints == 0:
            return None

        lb = optimization_problem.beq_transformed - optimization_problem.eps_lineq
        ub = optimization_problem.beq_transformed + optimization_problem.eps_lineq

        return optimize.LinearConstraint(
            optimization_problem.Aeq_independent_transformed, lb, ub, keep_feasible=True
        )

    def get_nonlincon_obj(self, optimization_problem: OptimizationProblem) -> list:
        """
        Return the optimized nonlinear constraints as an object.

        Returns
        -------
        nonlincon_obj : list
            Nonlinear constraint violation objects with bounds the optimization_problem.

        See Also
        --------
        constraint_objects
        nonlinear_constraints
        """
        if optimization_problem.n_nonlinear_constraints == 0:
            return [None]

        opt = optimization_problem

        def makeConstraint(i: int) -> optimize.NonlinearConstraint:
            """
            Create optimize.NonlinearConstraint object.

            Parameters
            ----------
            i : int
                Variable index

            Returns
            -------
            constr : optimize.NonlinearConstraint
                Constraint object.

            Notes
            -----
            Note, this is necessary to avoid side effects when creating the function
            in the main loop.
            """
            constr = optimize.NonlinearConstraint(
                lambda x: opt.evaluate_nonlinear_constraints_violation(
                    x,
                    untransform=True,
                    get_dependent_values=True,
                )[i],
                lb=-np.inf,
                ub=0,
                finite_diff_rel_step=self.finite_diff_rel_step,
                keep_feasible=True,
            )
            return constr

        constraints = []
        for i in range(opt.n_nonlinear_constraints):
            constraints.append(makeConstraint(i))

        return constraints

    def __str__(self) -> str:
        """str: String representation."""
        return self.__class__.__name__


class TrustConstr(SciPyInterface):
    """
    Wrapper for the trust-constr optimization method from the scipy optimization suite.

    It defines the solver options in the 'options' variable as a dictionary.

    Supports:
        - Linear constraints.
        - Linear equality constraints.
        - Nonlinear constraints.
        - Bounds.

    Parameters
    ----------
    gtol : UnsignedFloat, optional
        Tolerance for termination by the norm of the Lagrangian gradient.
        The algorithm will terminate when both the infinity norm (i.e., max abs value)
        of the Lagrangian gradient and the constraint violation are smaller than gtol.
        Default is 1e-8.
    xtol : UnsignedFloat, optional
        Tolerance for termination by the change of the independent variable.
        The algorithm will terminate when tr_radius < xtol,
        where tr_radius is the radius of the trust region used in the algorithm.
        Default is 1e-8.
    barrier_tol : UnsignedFloat, optional
        Threshold on the barrier parameter for the algorithm termination.
        When inequality constraints are present, the algorithm will terminate only
        when the barrier parameter is less than barrier_tol.
        Default is 1e-8.
    initial_tr_radius : float, optional
        Initial trust radius. The trust radius gives the maximum distance between solution points
        in consecutive iterations. It reflects the trust the algorithm puts in the local
        approximation of the optimization problem. For an accurate local approximation, the
        trust-region should be large, and for an approximation valid only close to the current
        point, it should be a small one. The trust radius is automatically updated throughout the
        optimization process, with initial_tr_radius being its initial value. Default is 1.
    initial_constr_penalty : float, optional
        Initial constraints penalty parameter. The penalty parameter is used for balancing
        the requirements of decreasing the objective function and satisfying the constraints.
        It is used for defining the merit function: merit_function(x) = fun(x)
        + constr_penalty * constr_norm_l2(x), where constr_norm_l2(x) is the l2 norm of a vector
        containing all the constraints. The merit function is used for accepting or rejecting trial
        points, and constr_penalty weights the two conflicting goals of reducing the objective
        function and constraints. The penalty is automatically updated throughout the optimization
        process, with initial_constr_penalty being its initial value. Default is 1.
    initial_barrier_parameter : float, optional
        Initial barrier parameter. Used only when inequality constraints are present.
        For dealing with optimization problems min_x f(x) subject to inequality constraints
        c(x) <= 0,
        the algorithm introduces slack variables, solving the problem
        min_(x, s) f(x) + barrier_parameter * sum(ln(s))
        subject to the equality constraints
        c(x) + s = 0
        instead of the original problem.
        This subproblem is solved for decreasing values of barrier_parameter and with decreasing
        tolerances for the termination, starting with initial_barrier_parameter for the barrier
        parameter. Default is 0.1.
    initial_barrier_tolerance : float, optional
        Initial tolerance for the barrier subproblem. Used only when inequality constraints are
        present. For dealing with optimization problems min_x f(x) subject to inequality constraints
        c(x) <= 0,
        the algorithm introduces slack variables, solving the problem
        min_(x, s) f(x) + barrier_parameter * sum(ln(s))
        subject to the equality constraints
        c(x) + s = 0
        instead of the original problem.
        This subproblem is solved for decreasing values of barrier_parameter
        and with decreasing tolerances for the termination, starting with initial_barrier_tolerance
        for the barrier tolerance. Default is 0.1.
    factorization_method : str or None, optional
        Method to factorize the Jacobian of the constraints.
        Use None (default) for auto selection or one of:
        - 'NormalEquation'
        - 'AugmentedSystem'
        - 'QRFactorization'
        - 'SVDFactorization'.
        The methods 'NormalEquation' and 'AugmentedSystem' can be used only with sparse
        constraints. The methods 'QRFactorization' and 'SVDFactorization' can be used
        only with dense constraints.
        Default is None.
    maxiter : UnsignedInteger, optional
        Maximum number of algorithm iterations. Default is 1000.
    verbose : UnsignedInteger, optional
        Level of algorithm's verbosity:
        - 0 (default) for silent
        - 1 for a termination report
        - 2 for progress during iterations
        - 3 for more complete progress report.
    disp : Bool, optional
        If True, then verbose will be set to 1 if it was 0. Default is False.
    """

    supports_linear_constraints = True
    supports_linear_equality_constraints = True
    supports_nonlinear_constraints = True
    supports_bounds = True

    gtol = UnsignedFloat(default=1e-8)
    xtol = UnsignedFloat(default=1e-8)
    barrier_tol = UnsignedFloat(default=1e-8)
    initial_tr_radius = UnsignedFloat(default=1.0)
    initial_constr_penalty = UnsignedFloat(default=1.0)
    initial_barrier_parameter = UnsignedFloat(default=0.1)
    initial_barrier_tolerance = UnsignedFloat(default=0.1)
    factorization_method = Switch(
        valid=[
            "NormalEquation",
            "AugmentedSystem",
            "QRFactorization",
            "SVDFactorization",
        ]
    )
    maxiter = UnsignedInteger(default=1000)
    verbose = UnsignedInteger(default=0)
    disp = Bool(default=False)

    x_tol = xtol            # Alias for uniform interface
    cv_nonlincon_tol = gtol  # Alias for uniform interface
    n_max_evals = maxiter   # Alias for uniform interface
    n_max_iter = maxiter    # Alias for uniform interface

    _specific_options = [
        "gtol",
        "xtol",
        "barrier_tol",
        "finite_diff_rel_step",
        "initial_constr_penalty",
        "initial_tr_radius",
        "initial_barrier_parameter",
        "initial_barrier_tolerance",
        "factorization_method",
        "maxiter",
        "verbose",
        "disp",
    ]

    def __str__(self) -> str:
        """str: String representation."""
        return "trust-constr"


class COBYLA(SciPyInterface):
    """
    Wrapper for the COBYLA optimization method from the scipy optimization suite.

    It defines the solver options in the 'options' variable as a dictionary.

    Supports:
        - Linear constraints
        - Linear equality constraints
        - Nonlinear constraints

    Parameters
    ----------
    rhobeg : float, default 1
        Reasonable initial changes to the variables.
    tol : float, default 0.0002
        Final accuracy in the optimization (not precisely guaranteed).
        This is a lower bound on the size of the trust region.
    disp : bool, default False
        Set to True to print convergence messages.
        If False, verbosity is ignored and set to 0.
    maxiter : int, default 10000
        Maximum number of function evaluations.
    catol : float, default 2e-4
        Absolute tolerance for constraint violations.
    """

    supports_linear_constraints = True
    supports_linear_equality_constraints = True
    supports_nonlinear_constraints = True
    supports_bounds = True

    rhobeg = UnsignedFloat(default=1)
    tol = UnsignedFloat(default=0.0002)
    maxiter = UnsignedInteger(default=10000)
    disp = Bool(default=False)
    catol = UnsignedFloat(default=0.0002)

    x_tol = tol                 # Alias for uniform interface
    cv_nonlincon_tol = catol    # Alias for uniform interface
    n_max_evals = maxiter       # Alias for uniform interface
    n_max_iter = maxiter        # Alias for uniform interface

    _specific_options = ["rhobeg", "tol", "maxiter", "disp", "catol"]


class NelderMead(SciPyInterface):
    """
    Wrapper for the Nelder-Mead optimization method from the scipy optimization suite.

    Supports:
        - Bounds.

    It defines the solver options in the 'options' variable as a dictionary.

    Parameters
    ----------
    maxiter : UnsignedInteger
        Maximum allowed number of iterations. The default = 1000.
    initial_simplex : None or array_like, optional
        Initial simplex. If given, it overrides x0.
        initial_simplex[j, :] should contain the coordinates of the jth vertex of the
        N+1 vertices in the simplex, where N is the dimension.
    xatol : UnsignedFloat, optional
        Absolute error in xopt between iterations that is acceptable for convergence.
    fatol : UnsignedFloat, optional
        Absolute error in f(xopt) between iterations that is acceptable for convergence.
    adaptive : Bool, optional
        Adapt algorithm parameters to dimensionality of the problem.
        Useful for high-dimensional minimization.
    disp : Bool, optional
        Set to True to print convergence messages.
    """

    supports_bounds = True

    maxiter = UnsignedInteger(default=1000)
    initial_simplex = None
    xatol = UnsignedFloat(default=1e-3)
    fatol = UnsignedFloat(default=1e-3)
    adaptive = Bool(default=True)
    disp = Bool(default=False)

    x_tol = xatol           # Alias for uniform interface
    f_tol = fatol           # Alias for uniform interface
    n_max_evals = maxiter   # Alias for uniform interface
    n_max_iter = maxiter    # Alias for uniform interface

    _specific_options = [
        "maxiter",
        "initial_simplex",
        "xatol",
        "fatol",
        "adaptive",
        "disp",
    ]

    def __str__(self) -> str:
        """str: String representation."""
        return "Nelder-Mead"


class SLSQP(SciPyInterface):
    """
    Wrapper for the SLSQP optimization method from the scipy optimization suite.

    It defines the solver options in the 'options' variable as a dictionary.

    Supports:
        - Linear constraints
        - Linear equality constraints
        - Nonlinear constraints
        - Bounds

    Parameters
    ----------
    ftol : float, default 1e-2
        Precision goal for the value of f in the stopping criterion.
    eps : float, default 1e-6
        Step size used for numerical approximation of the Jacobian.
    disp : bool, default False
        Set to True to print convergence messages.
        If False, verbosity is ignored and set to 0.
    maxiter : int, default 1000
        Maximum number of iterations.
    iprint: int, optional
        The verbosity of fmin_slsqp :
            iprint <= 0 : Silent operation
            iprint == 1 : Print summary upon completion (default)
            iprint >= 2 : Print status of each iterate and summary
    """

    supports_linear_constraints = True
    supports_linear_equality_constraints = True
    supports_nonlinear_constraints = True
    supports_bounds = True

    ftol = UnsignedFloat(default=1e-2)
    eps = UnsignedFloat(default=1e-6)
    disp = Bool(default=False)
    maxiter = UnsignedInteger(default=1000)
    iprint = UnsignedInteger(ub=2, default=1)

    f_tol = ftol            # Alias for uniform interface
    n_max_evals = maxiter   # Alias for uniform interface
    n_max_iter = maxiter    # Alias for uniform interface

    _specific_options = [
        "ftol",
        "eps",
        "disp",
        "maxiter",
        "finite_diff_rel_step",
        "iprint",
    ]



"""
================================================
Process Model (:mod:`CADETProcess.processModel`)
================================================

.. currentmodule:: CADETProcess.processModel

Classes for modelling processes.

ComponentSystem
===============

.. autosummary::
    :toctree: generated/

    Species
    Component
    ComponentSystem

Reaction Models
===============

.. autosummary::
    :toctree: generated/

    Reaction
    CrossPhaseReaction
    ReactionBaseClass
    NoReaction
    MassActionLaw
    MassActionLawParticle

Binding Models
==============

.. autosummary::
    :toctree: generated/

    BindingBaseClass
    NoBinding
    Linear
    Langmuir
    LangmuirLDF
    LangmuirLDFLiquidPhase
    BiLangmuir
    BiLangmuirLDF
    FreundlichLDF
    StericMassAction
    AntiLangmuir
    Spreading
    MobilePhaseModulator
    ExtendedMobilePhaseModulator
    SelfAssociation
    BiStericMassAction
    MultistateStericMassAction
    SimplifiedMultistateStericMassAction
    Saska
    GeneralizedIonExchange
    HICConstantWaterActivity
    HICWaterOnHydrophobicSurfaces
    MultiComponentColloidal

Unit Operation Models
=====================

.. autosummary::
    :toctree: generated/

    UnitBaseClass
    Inlet
    Outlet
    Cstr
    TubularReactor
    LumpedRateModelWithoutPores
    LumpedRateModelWithPores
    GeneralRateModel
    MCT

Discretization
--------------

.. autosummary::
    :toctree: generated/

    discretization

Solution Recorder
-----------------

.. autosummary::
    :toctree: generated/

    solutionRecorder

Flow Sheet
==========

.. autosummary::
    :toctree: generated/

    FlowSheet

Process
=======

.. autosummary::
    :toctree: generated/

    Process

"""

from . import componentSystem
from .componentSystem import *

from . import reaction
from .reaction import *

from . import binding
from .binding import *

from . import discretization
from .discretization import *

from . import unitOperation
from .unitOperation import *

from . import flowSheet
from .flowSheet import *

from . import process
from .process import *

__all__ = [s for s in dir()]



from typing import Optional, Sequence

import numpy as np

from CADETProcess.processModel import MassActionLaw

from . import ptc


def calculate_buffer_equilibrium(
    buffer: Sequence[float],
    reaction_system: MassActionLaw,
    constant_indices: Optional[Sequence[int]] = None,
    reinit: bool = True,
    verbose: bool = False,
) -> list[float]:
    """
    Calculate buffer equilibrium for given concentration.

    Parameters
    ----------
    buffer : list of floats
        Buffer concentration in mM
    reaction_system : MassActionLaw
        Reaction rates and stoichiometric matrix for calculating equilibrium.
    constant_indices : list, optional
        Indices of fixed target concentration (e.g. proton concentration/pH).
    reinit: Bool, optional
        If True, run CADET with initial values to get 'smooth' initial values
    verbose : Bool, optional
        If True, print information at every ptc iteration.

    Returns
    -------
    sol : list of floats.
        Buffer equilbrium concentrations
    """

    def residual(c: np.ndarray) -> np.ndarray:
        return dydx_mal(
            c,
            reaction_system,
            constant_indices,
            buffer.copy(),
        )

    def jacobian(c: np.ndarray) -> np.ndarray:
        return jac_mal(
            c,
            reaction_system,
            constant_indices,
            buffer.copy(),
        )

    stats, sol, res, k = ptc(
        np.array(buffer.copy()),
        residual,
        jacobian,
        1e-4,   # init step size
        1e-14,  # tolerance (scaled l2)
        quiet=not (verbose),
        maxIter=10000,
    )

    return np.round(np.abs(sol), 14).tolist()


def dydx_mal(
    c: np.ndarray,
    reaction_system: MassActionLaw,
    constant_indices: Optional[Sequence[int]] = None,
    c_init: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Compute the time derivative of concentrations in a mass action law system.

    Parameters
    ----------
    c : np.ndarray
        Concentration vector of components.
    reaction_system : MassActionLaw
        Reaction rates and stoichiometric matrix for calculating equilibrium.
    constant_indices : Optional[Sequence[int]]
        Indices of components whose concentrations are to be held constant.
    c_init : Optional[np.ndarray]
        Initial concentration vector (used to fix constants if constant_indices is provided).

    Returns
    -------
    dydx : np.ndarray
        Time derivative of concentration vector.
    """
    cc = np.asarray(c, dtype="float64")
    if constant_indices is not None:
        if c_init is None:
            c_init = c
        for comp in constant_indices:
            cc[comp] = c_init[comp]

    exp_fwd = reaction_system.exponents_fwd
    exp_bwd = reaction_system.exponents_bwd
    k_fwd = reaction_system.k_fwd
    k_bwd = reaction_system.k_bwd

    r = np.zeros(reaction_system.n_reactions)

    for r_i in range(reaction_system.n_reactions):
        fwd_indices = np.where(exp_fwd[:, r_i] > 0.0)
        prod = np.prod(cc[fwd_indices] ** exp_fwd[:, r_i][fwd_indices])
        r_fwd = k_fwd[r_i] * prod

        bwd_indices = np.where(exp_bwd[:, r_i] > 0.0)
        prod = np.prod(cc[bwd_indices] ** exp_bwd[:, r_i][bwd_indices])
        r_bwd = k_bwd[r_i] * prod

        r[r_i] = r_fwd - r_bwd

    dydx = np.dot(reaction_system.stoich, r)

    if constant_indices is not None:
        for comp in constant_indices:
            dydx[comp] = 0

    return dydx


def jac_mal(
    c: np.ndarray,
    reaction_system: MassActionLaw,
    constant_indices: Optional[Sequence[int]] = None,
    c_init: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Compute the Jacobian of a mass action law reaction system at given concentrations.

    Parameters
    ----------
    c : np.ndarray
        Current concentrations of components.
    reaction_system : MassActionLaw
        Reaction system object containing stoichiometry, exponents, and rate constants.
    constant_indices : Optional[Sequence[int]]
        Indices of components to be treated as constant
        (i.e., their derivatives are zeroed out).
    c_init : Optional[np.ndarray]
        Initial concentration vector to reset constants if `constant_indices` is provided.

    Returns
    -------
    jac : np.ndarray
        Jacobian matrix (n_comp x n_comp) of the rate equations.

    """
    cc = np.asarray(c, dtype="float64")

    if constant_indices is not None:
        if c_init is None:
            c_init = c
        for comp in constant_indices:
            cc[comp] = c_init[comp]

    exp_fwd = reaction_system.exponents_fwd
    exp_bwd = reaction_system.exponents_bwd
    k_fwd = reaction_system.k_fwd
    k_bwd = reaction_system.k_bwd

    jac_r = np.zeros((reaction_system.n_reactions, reaction_system.n_comp))

    for r_i in range(reaction_system.n_reactions):
        fwd_indices = np.where(exp_fwd[:, r_i] > 0.0)
        j_fwd = np.zeros_like(cc)
        j_fwd[fwd_indices] = k_fwd[r_i]

        bwd_indices = np.where(exp_bwd[:, r_i] > 0.0)
        j_bwd = np.zeros_like(cc)
        j_bwd[bwd_indices] = k_bwd[r_i]

        for comp in range(len(c)):
            exponents_fwd = exp_fwd[:, r_i]
            exponents_bwd = exp_bwd[:, r_i]
            if exponents_fwd[comp] > 0.0:
                exp_derivative = exponents_fwd.copy()
                exp_derivative[comp] -= 1.0
                prod = np.prod(cc**exp_derivative)
                j_fwd[comp] *= exponents_fwd[comp] * prod
            if exponents_bwd[comp] > 0.0:
                exp_derivative = exponents_bwd.copy()
                exp_derivative[comp] -= 1.0
                prod = np.prod(cc**exp_derivative)
                j_bwd[comp] *= exponents_bwd[comp] * prod

        jac_r[r_i, :] = j_fwd - j_bwd

    jac = np.dot(reaction_system.stoich, jac_r)

    if constant_indices is not None:
        for comp in constant_indices:
            jac[comp, :] = 0

    return jac



import unittest

import numpy as np
from CADETProcess import CADETProcessError
from CADETProcess.modelBuilder import CompartmentBuilder
from CADETProcess.processModel import ComponentSystem, Langmuir


class Test_CompartmentBuilder(unittest.TestCase):
    def setUp(self):
        self.component_system = ComponentSystem(2)

        self.binding_model = Langmuir(self.component_system)

        self.volume_simple = [1, 2, 3, 4, 5]
        self.matrix_simple = [
            0,   0.1, 0.2, 0.3, 0.4,  # noqa: E241
            0.1, 0,   0,   0,   0,    # noqa: E241
            0.2, 0,   0,   0,   0,    # noqa: E241
            0.3, 0,   0,   0,   0,    # noqa: E241
            0.4, 0,   0,   0,   0,    # noqa: E241
        ]

        self.builder_simple = CompartmentBuilder(
            self.component_system,
            self.volume_simple,
            self.matrix_simple,
        )
        self.builder_simple.cycle_time = 1000

        self.volume_complex = ["inlet", 2, 1, 3, 1, 2, 1, 4, 1, "outlet"]
        # fmt: off
        self.matrix_complex = [
        #   0    1    2    3    4    5    6    7    8    9
            0,   0.1, 0,   0,   0,   0,   0,   0,   0,   0,    # 0 #noqa: E241
            0,   0,   0.3, 0,   0,   0,   0,   0.1, 0,   0,    # 1 #noqa: E241
            0,   0.1, 0,   0.1, 0,   0,   0,   0.1, 0,   0,    # 2 #noqa: E241
            0,   0.2, 0,   0,   0,   0.5, 0,   0,   0.1, 0,    # 3 #noqa: E241
            0,   0,   0,   0.1, 0,   0.1, 0,   0.1, 0,   0,    # 4 #noqa: E241
            0,   0,   0,   0,   0.3, 0,   0.2, 0.1, 0,   0,    # 5 #noqa: E241
            0,   0,   0,   0,   0,   0,   0,   0.1, 0,   0.1,  # 6 #noqa: E241
            0,   0,   0,   0.5, 0,   0,   0,   0,   0,   0,    # 7 #noqa: E241
            0,   0,   0,   0.1, 0,   0,   0,   0,   0,   0,    # 8 #noqa: E241
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,    # 9 #noqa: E241
        ]
        # fmt: on

        self.builder_complex = CompartmentBuilder(
            self.component_system,
            self.volume_complex,
            self.matrix_complex,
            init_c=1,
        )
        self.builder_complex.cycle_time = 1000

    def test_binding(self):
        self.builder_simple.binding_model = self.binding_model
        self.assertIsInstance(self.builder_simple.binding_model, Langmuir)
        compartment = self.builder_simple.flow_sheet.compartment_2
        self.assertIsInstance(compartment.binding_model, Langmuir)

    def test_complex(self):
        self.builder_complex.validate_flow_rates()

    def test_connections(self):
        flow_rates_expected = {
            "compartment_0": {
                "total_in": {None: np.array([1.0, 0.0, 0.0, 0.0])},
                "total_out": {None: np.array([1.0, 0.0, 0.0, 0.0])},
                "origins": {
                    None: {
                        "compartment_1": {None: np.array([0.1, 0.0, 0.0, 0.0])},
                        "compartment_2": {None: np.array([0.2, 0.0, 0.0, 0.0])},
                        "compartment_3": {None: np.array([0.3, 0.0, 0.0, 0.0])},
                        "compartment_4": {None: np.array([0.4, 0.0, 0.0, 0.0])},
                    }
                },
                "destinations": {
                    None: {
                        "compartment_1": {None: np.array([0.1, 0.0, 0.0, 0.0])},
                        "compartment_2": {None: np.array([0.2, 0.0, 0.0, 0.0])},
                        "compartment_3": {None: np.array([0.3, 0.0, 0.0, 0.0])},
                        "compartment_4": {None: np.array([0.4, 0.0, 0.0, 0.0])},
                    }
                },
            },
            "compartment_1": {
                "total_in": {None: np.array([0.1, 0.0, 0.0, 0.0])},
                "total_out": {None: np.array([0.1, 0.0, 0.0, 0.0])},
                "origins": {
                    None: {"compartment_0": {None: np.array([0.1, 0.0, 0.0, 0.0])}}
                },
                "destinations": {
                    None: {"compartment_0": {None: np.array([0.1, 0.0, 0.0, 0.0])}}
                },
            },
            "compartment_2": {
                "total_in": {None: np.array([0.2, 0.0, 0.0, 0.0])},
                "total_out": {None: np.array([0.2, 0.0, 0.0, 0.0])},
                "origins": {
                    None: {"compartment_0": {None: np.array([0.2, 0.0, 0.0, 0.0])}}
                },
                "destinations": {
                    None: {"compartment_0": {None: np.array([0.2, 0.0, 0.0, 0.0])}}
                },
            },
            "compartment_3": {
                "total_in": {None: np.array([0.3, 0.0, 0.0, 0.0])},
                "total_out": {None: np.array([0.3, 0.0, 0.0, 0.0])},
                "origins": {
                    None: {"compartment_0": {None: np.array([0.3, 0.0, 0.0, 0.0])}}
                },
                "destinations": {
                    None: {"compartment_0": {None: np.array([0.3, 0.0, 0.0, 0.0])}}
                },
            },
            "compartment_4": {
                "total_in": {None: np.array([0.4, 0.0, 0.0, 0.0])},
                "total_out": {None: np.array([0.4, 0.0, 0.0, 0.0])},
                "origins": {
                    None: {"compartment_0": {None: np.array([0.4, 0.0, 0.0, 0.0])}}
                },
                "destinations": {
                    None: {"compartment_0": {None: np.array([0.4, 0.0, 0.0, 0.0])}}
                },
            },
        }
        flow_rates = self.builder_simple.flow_sheet.get_flow_rates().to_dict()
        np.testing.assert_equal(flow_rates, flow_rates_expected)

    def test_validate_flow_rates(self):
        self.builder_complex.flow_sheet.compartment_2.flow_rate = 0

        with self.assertRaises(CADETProcessError):
            self.builder_complex.validate_flow_rates()

    def test_initial_conditions(self):
        # All to zero (default)
        builder = CompartmentBuilder(
            self.component_system,
            self.volume_complex,
            self.matrix_complex,
        )
        c_expected = [0, 0]
        c = builder.flow_sheet.compartment_2.c
        np.testing.assert_almost_equal(c, c_expected)

        # All components, all compartments same value
        builder = CompartmentBuilder(
            self.component_system,
            self.volume_complex,
            self.matrix_complex,
            init_c=1,
        )
        c_expected = [1, 1]
        c = builder.flow_sheet.compartment_2.c
        np.testing.assert_almost_equal(c, c_expected)

        # All compartments same value
        builder = CompartmentBuilder(
            self.component_system,
            self.volume_complex,
            self.matrix_complex,
            init_c=[1, 2],
        )
        c_expected = [1, 2]
        c = builder.flow_sheet.compartment_2.c
        np.testing.assert_almost_equal(c, c_expected)

        # Different per zone
        init_c = np.array([[0, 0], [1, 1], [2, 2], [3, 3], [4, 4]])

        builder = CompartmentBuilder(
            self.component_system,
            self.volume_simple,
            self.matrix_simple,
            init_c=init_c,
        )
        c = builder.flow_sheet.compartment_2.c
        np.testing.assert_almost_equal(c, init_c[2].tolist())

    def test_tracer(self):
        """
        ..todo::
            flow rate validation not working for flow rate filter
        """
        self.builder_complex.add_tracer(4, [1, 1], 10, 0.1)

    def test_process(self):
        self.builder_complex.cycle_time = 1000

        self.builder_complex.add_tracer(4, [1, 1], 10, 0.1)

        from CADETProcess.simulator import Cadet

        process_simulator = Cadet()

        proc_results = process_simulator.simulate(self.builder_complex.process)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from CADETProcess.processModel import (
    MCT,
    ComponentSystem,
    Cstr,
    FlowSheet,
    GeneralRateModel,
    Inlet,
    LumpedRateModelWithoutPores,
    LumpedRateModelWithPores,
    Outlet,
    Process,
    StericMassAction,
    TubularReactor,
)


def create_lwe(unit_type: str = "GeneralRateModel", **kwargs) -> Process:
    """
    Create a process with the specified unit type and configuration.

    Parameters
    ----------
    unit_type : str, optional
        The type of unit operation, by default 'GeneralRateModel'.
    **kwargs : dict
        Additional parameters for configuring the unit operation.

    Returns
    -------
    Process
        The configured process.
    """
    n_comp: int = kwargs.get("n_comp", 4)
    component_system = ComponentSystem(n_comp)

    if unit_type == "Cstr":
        unit = configure_cstr(component_system, **kwargs)
    elif unit_type == "GeneralRateModel":
        unit = configure_general_rate_model(component_system, **kwargs)
    elif unit_type == "TubularReactor":
        unit = configure_tubular_reactor(component_system, **kwargs)
    elif unit_type == "LumpedRateModelWithoutPores":
        unit = configure_lumped_rate_model_without_pores(component_system, **kwargs)
    elif unit_type == "LumpedRateModelWithPores":
        unit = configure_lumped_rate_model_with_pores(component_system, **kwargs)
    elif unit_type == "MCT":
        unit = configure_multichannel_transport_model(component_system, **kwargs)
    else:
        raise ValueError(f"Unknown unit operation type {unit_type}")

    flow_sheet = setup_flow_sheet(unit, component_system)

    process = Process(flow_sheet, "process")
    process.cycle_time = 120 * 60

    c1_lwe = [[50.0], [0.0], [[100.0, 0.2]]]
    cx_lwe = [[1.0], [0.0], [0.0]]

    process.add_event(
        "load", "flow_sheet.inlet.c", c1_lwe[0] + cx_lwe[0] * (n_comp - 1), 0
    )
    process.add_event(
        "wash", "flow_sheet.inlet.c", c1_lwe[1] + cx_lwe[1] * (n_comp - 1), 10
    )
    process.add_event(
        "elute", "flow_sheet.inlet.c", c1_lwe[2] + cx_lwe[2] * (n_comp - 1), 90
    )

    return process


def setup_flow_sheet(unit, component_system: ComponentSystem) -> FlowSheet:
    """
    Set up the flow sheet for the process.

    Parameters
    ----------
    unit : UnitOperation
        The unit operation to be added to the flow sheet.
    component_system : ComponentSystem
        The component system of the process.

    Returns
    -------
    FlowSheet
        The configured flow sheet.
    """
    flow_sheet = FlowSheet(component_system)
    inlet = Inlet(component_system, name="inlet")
    inlet.flow_rate = 1.2e-3
    outlet = Outlet(component_system, name="outlet")

    flow_sheet.add_unit(inlet)
    flow_sheet.add_unit(unit)
    flow_sheet.add_unit(outlet)

    if unit.has_ports:
        flow_sheet.add_connection(inlet, unit, destination_port="channel_0")
        flow_sheet.add_connection(unit, outlet, origin_port="channel_0")
    else:
        flow_sheet.add_connection(inlet, unit)
        flow_sheet.add_connection(unit, outlet)

    return flow_sheet


def configure_cstr(component_system: ComponentSystem, **kwargs) -> Cstr:
    """
    Configure a continuous stirred-tank reactor (CSTR).

    Parameters
    ----------
    component_system : ComponentSystem
        The component system of the process.
    **kwargs : dict
        Additional parameters for configuring the CSTR.

    Returns
    -------
    Cstr
        The configured CSTR.
    """
    cstr = Cstr(component_system, name="Cstr")

    total_volume = 1e-3
    total_porosity = 0.37 + (1.0 - 0.37) * 0.75
    cstr.init_liquid_volume = total_porosity * total_volume
    cstr.const_solid_volume = (1 - total_porosity) * total_volume

    configure_solution_recorder(cstr, **kwargs)
    configure_steric_mass_action(cstr, component_system, **kwargs)

    return cstr


def configure_general_rate_model(
    component_system: ComponentSystem, **kwargs
) -> GeneralRateModel:
    """
    Configure a general rate model.

    Parameters
    ----------
    component_system : ComponentSystem
        The component system of the process.
    **kwargs : dict
        Additional parameters for configuring the general rate model.

    Returns
    -------
    GeneralRateModel
        The configured general rate model.
    """
    grm = GeneralRateModel(component_system, name="GeneralRateModel")

    grm.length = 0.014
    grm.diameter = 0.01 * 2
    grm.bed_porosity = 0.37
    grm.axial_dispersion = 5.75e-8
    grm.pore_diffusion = [7e-10, 6.07e-11, 6.07e-11, 6.07e-11]

    configure_solution_recorder(grm, **kwargs)
    configure_discretization(grm, **kwargs)
    configure_particles(grm)
    configure_steric_mass_action(grm, component_system, **kwargs)
    configure_film_diffusion(grm, component_system.n_comp)
    configure_flow_direction(grm, **kwargs)

    return grm


def configure_tubular_reactor(
    component_system: ComponentSystem, **kwargs
) -> TubularReactor:
    """
    Configure a tubular reactor.

    Parameters
    ----------
    component_system : ComponentSystem
        The component system of the process.
    **kwargs : dict
        Additional parameters for configuring the tubular reactor.

    Returns
    -------
    TubularReactor
        The configured tubular reactor.
    """
    tr = TubularReactor(component_system, name="TubularReactor")

    tr.length = 0.014
    tr.diameter = 0.01 * 2
    tr.axial_dispersion = 5.75e-8

    configure_solution_recorder(tr, **kwargs)
    configure_discretization(tr, **kwargs)
    configure_flow_direction(tr, **kwargs)

    return tr


def configure_lumped_rate_model_without_pores(
    component_system: ComponentSystem, **kwargs
) -> LumpedRateModelWithoutPores:
    """
    Configure a lumped rate model without pores.

    Parameters
    ----------
    component_system : ComponentSystem
        The component system of the process.
    **kwargs : dict
        Additional parameters for configuring the lumped rate model.

    Returns
    -------
    LumpedRateModelWithoutPores
        The configured lumped rate model.
    """
    lrm = LumpedRateModelWithoutPores(
        component_system, name="LumpedRateModelWithoutPores"
    )

    lrm.length = 0.014
    lrm.diameter = 0.01 * 2
    lrm.total_porosity = 0.37 + (1.0 - 0.37) * 0.75
    lrm.axial_dispersion = 5.75e-8

    configure_solution_recorder(lrm, **kwargs)
    configure_discretization(lrm, **kwargs)
    configure_steric_mass_action(lrm, component_system, **kwargs)
    configure_flow_direction(lrm, **kwargs)

    return lrm


def configure_lumped_rate_model_with_pores(
    component_system: ComponentSystem, **kwargs
) -> LumpedRateModelWithPores:
    """
    Configure a lumped rate model with pores.

    Parameters
    ----------
    component_system : ComponentSystem
        The component system of the process.
    **kwargs : dict
        Additional parameters for configuring the lumped rate model.

    Returns
    -------
    LumpedRateModelWithPores
        The configured lumped rate model.
    """
    lrmp = LumpedRateModelWithPores(component_system, name="LumpedRateModelWithPores")

    lrmp.length = 0.014
    lrmp.diameter = 0.01 * 2
    lrmp.bed_porosity = 0.37
    lrmp.axial_dispersion = 5.75e-8

    configure_solution_recorder(lrmp, **kwargs)
    configure_discretization(lrmp, **kwargs)
    configure_particles(lrmp)
    configure_steric_mass_action(lrmp, component_system, **kwargs)
    configure_film_diffusion(lrmp, component_system.n_comp)
    configure_flow_direction(lrmp, **kwargs)

    return lrmp


def configure_multichannel_transport_model(
    component_system: ComponentSystem, **kwargs
) -> MCT:
    """
    Configure a multichannel transport model.

    Parameters
    ----------
    component_system : ComponentSystem
        The component system of the process.
    **kwargs : dict
        Additional parameters for configuring the multichannel transport model.

    Returns
    -------
    MCT
        The configured multichannel transport model.
    """
    mct = MCT(component_system, nchannel=3, name="MCT")

    mct.length = 0.014
    mct.channel_cross_section_areas = 3 * [2 * np.pi * (0.01**2)]
    mct.axial_dispersion = 5.75e-8

    n_comp: int = component_system.n_comp

    mct.exchange_matrix = np.array(
        [
            [n_comp * [0.0], n_comp * [0.001], n_comp * [0.0]],
            [n_comp * [0.002], n_comp * [0.0], n_comp * [0.003]],
            [n_comp * [0.0], n_comp * [0.0], n_comp * [0.0]],
        ]
    )

    configure_solution_recorder(mct, **kwargs)
    configure_discretization(mct, **kwargs)
    configure_flow_direction(mct, **kwargs)

    return mct


def configure_discretization(unit_operation, **kwargs) -> None:
    """
    Configure the discretization settings for a unit operation.

    Parameters
    ----------
    unit_operation : UnitOperation
        The unit operation to be configured.
    **kwargs : dict
        Additional parameters for configuring the discretization.
    """
    n_col: int = kwargs.get("n_col", 100)
    n_par: int = kwargs.get("n_par", 2)
    ad_jacobian: bool = kwargs.get("ad_jacobian", False)

    if "npar" in unit_operation.discretization.parameters:
        unit_operation.discretization.npar = n_par
        unit_operation.discretization.par_disc_type = "EQUIDISTANT_PAR"

    unit_operation.discretization.ncol = n_col
    unit_operation.discretization.use_analytic_jacobian = not ad_jacobian


def configure_particles(unit_operation) -> None:
    """
    Configure the particle settings for a unit operation.

    Parameters
    ----------
    unit_operation : UnitOperation
        The unit operation to be configured.
    """
    par_radius = 4.5e-5
    par_porosity = 0.75

    unit_operation.particle_radius = par_radius
    unit_operation.particle_porosity = par_porosity
    unit_operation.discretization.par_geom = "SPHERE"


def configure_steric_mass_action(unit_operation, component_system, **kwargs) -> None:
    """
    Configure the steric mass action binding model for a unit operation.

    Parameters
    ----------
    unit_operation : UnitOperation
        The unit operation to be configured.
    component_system : ComponentSystem
        The component system of the process.
    **kwargs : dict
        Additional parameters for configuring the steric mass action binding model.
    """
    is_kinetic = kwargs.get("is_kinetic", True)

    kA = 35.5
    kD = 1000.0
    nu = 4.7
    sigma = 11.83
    sma_lambda = 1.2e3

    binding_model = StericMassAction(component_system)

    binding_model.is_kinetic = is_kinetic
    binding_model.n_binding_sites = 1
    binding_model.adsorption_rate = kA
    binding_model.desorption_rate = kD
    binding_model.characteristic_charge = nu
    binding_model.steric_factor = sigma
    binding_model.capacity = sma_lambda

    unit_operation.binding_model = binding_model


def configure_film_diffusion(unit_operation, n_comp) -> None:
    """
    Configure the film diffusion settings for a unit operation.

    Parameters
    ----------
    unit_operation : UnitOperation
        The unit operation to be configured.
    n_comp : int
        The number of components in the process.
    """
    unit_operation.film_diffusion = [6.9e-6] * n_comp


def configure_flow_direction(unit_operation, **kwargs) -> None:
    """
    Configure the flow direction for a unit operation.

    Parameters
    ----------
    unit_operation : UnitOperation
        The unit operation to be configured.
    **kwargs : dict
        Additional parameters for configuring the flow direction.
    """
    reverse_flow = kwargs.get("reverse_flow", False)
    unit_operation.flow_direction = -1 if reverse_flow else 1


def configure_solution_recorder(unit_operation, **kwargs) -> None:
    """
    Configure the solution recorder for a unit operation.

    Parameters
    ----------
    unit_operation : UnitOperation
        The unit operation to be configured.
    **kwargs : dict
        Additional parameters for configuring the solution recorder.
    """
    for write_solution, value in unit_operation.solution_recorder.parameters.items():
        if value is False:
            unit_operation.solution_recorder.parameters[write_solution] = True



import unittest

import numpy as np
from CADETProcess.performance import Performance, RankedPerformance


class PerformanceTestCase(unittest.TestCase):
    def setUp(self):
        # Create example performance data
        mass = np.array([10.0, 20.0, 30.0])
        concentration = np.array([0.5, 0.4, 0.3])
        purity = np.array([0.9, 0.85, 0.95])
        recovery = np.array([0.8, 0.75, 0.9])
        productivity = np.array([0.7, 0.6, 0.65])
        eluent_consumption = np.array([1.5, 2.0, 1.2])
        mass_balance_difference = np.array([0.2, 0.1, -0.1])

        self.performance = Performance(
            mass,
            concentration,
            purity,
            recovery,
            productivity,
            eluent_consumption,
            mass_balance_difference,
        )

    def test_attributes(self):
        self.assertEqual(self.performance.n_comp, 3)

        np.testing.assert_array_equal(
            self.performance.mass, np.array([10.0, 20.0, 30.0])
        )
        np.testing.assert_array_equal(
            self.performance.concentration, np.array([0.5, 0.4, 0.3])
        )
        np.testing.assert_array_equal(
            self.performance.purity, np.array([0.9, 0.85, 0.95])
        )
        np.testing.assert_array_equal(
            self.performance.recovery, np.array([0.8, 0.75, 0.9])
        )
        np.testing.assert_array_equal(
            self.performance.productivity, np.array([0.7, 0.6, 0.65])
        )
        np.testing.assert_array_equal(
            self.performance.eluent_consumption, np.array([1.5, 2.0, 1.2])
        )
        np.testing.assert_array_equal(
            self.performance.mass_balance_difference, np.array([0.2, 0.1, -0.1])
        )

    def test_to_dict(self):
        expected_dict = {
            "mass": [10.0, 20.0, 30.0],
            "concentration": [0.5, 0.4, 0.3],
            "purity": [0.9, 0.85, 0.95],
            "recovery": [0.8, 0.75, 0.9],
            "productivity": [0.7, 0.6, 0.65],
            "eluent_consumption": [1.5, 2.0, 1.2],
            "mass_balance_difference": [0.2, 0.1, -0.1],
        }
        self.assertDictEqual(self.performance.to_dict(), expected_dict)


class RankedPerformanceTestCase(unittest.TestCase):
    def setUp(self):
        # Create example performance data
        mass = np.array([10.0, 20.0, 30.0])
        concentration = np.array([0.5, 0.4, 0.3])
        purity = np.array([0.9, 0.85, 0.95])
        recovery = np.array([0.8, 0.75, 0.9])
        productivity = np.array([0.7, 0.6, 0.65])
        eluent_consumption = np.array([1.5, 2.0, 1.2])
        mass_balance_difference = np.array([0.2, 0.1, -0.1])

        performance = Performance(
            mass,
            concentration,
            purity,
            recovery,
            productivity,
            eluent_consumption,
            mass_balance_difference,
        )

        self.ranked_performance_equal = RankedPerformance(performance, ranking=1)
        self.ranked_performance_diff = RankedPerformance(performance, ranking=[0, 1, 2])

    def test_attributes(self):
        np.testing.assert_array_equal(self.ranked_performance_equal.mass, 20)
        np.testing.assert_array_equal(
            self.ranked_performance_diff.mass, 26.666666666666668
        )
        np.testing.assert_array_equal(
            self.ranked_performance_equal.concentration, 0.39999999999999997
        )
        np.testing.assert_array_equal(
            self.ranked_performance_diff.concentration, 0.3333333333333333
        )
        np.testing.assert_array_equal(self.ranked_performance_equal.purity, 0.9)
        np.testing.assert_array_equal(
            self.ranked_performance_diff.purity, 0.9166666666666666
        )
        np.testing.assert_array_equal(
            self.ranked_performance_equal.recovery, 0.8166666666666668
        )
        np.testing.assert_array_equal(self.ranked_performance_diff.recovery, 0.85)
        np.testing.assert_array_equal(
            self.ranked_performance_equal.productivity, 0.6499999999999999
        )
        np.testing.assert_array_equal(
            self.ranked_performance_diff.productivity, 0.6333333333333333
        )
        np.testing.assert_array_equal(
            self.ranked_performance_equal.eluent_consumption, 1.5666666666666667
        )
        np.testing.assert_array_equal(
            self.ranked_performance_diff.eluent_consumption, 1.4666666666666668
        )
        np.testing.assert_array_equal(
            self.ranked_performance_equal.mass_balance_difference, 0.06666666666666668
        )
        np.testing.assert_array_equal(
            self.ranked_performance_diff.mass_balance_difference, -0.03333333333333333
        )


if __name__ == "__main__":
    unittest.main()



import os
from collections import defaultdict
from functools import wraps
from typing import Any, Callable, Optional

import numpy as np
from addict import Dict
from matplotlib.axes import Axes

from CADETProcess import CADETProcessError, SimulationResults, plotting, settings
from CADETProcess.dataStructure import String
from CADETProcess.dynamicEvents import Event, EventHandler
from CADETProcess.fractionation.fractions import Fraction, FractionPool
from CADETProcess.performance import Performance
from CADETProcess.processModel import ComponentSystem, Process
from CADETProcess.solution import SolutionIO, slice_solution

__all__ = ["Fractionator"]


class Fractionator(EventHandler):
    """
    Class for Chromatogram Fractionation.

    This class is responsible for setting events for starting and ending fractionation,
    handling multiple chromatograms, and calculating various performance metrics.

    Attributes
    ----------
    name : String
        Name of the fractionator, defaulting to 'Fractionator'.
    performance_keys : list
        Keys for performance metrics including mass, concentration, purity, recovery,
        productivity, and eluent consumption.
    """

    name = String(default="Fractionator")
    performance_keys: list[str] = [
        "mass",
        "concentration",
        "purity",
        "recovery",
        "productivity",
        "eluent_consumption",
    ]

    def __init__(
        self,
        simulation_results: SimulationResults,
        components: Optional[list[str]] = None,
        use_total_concentration_components: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the Fractionator.

        Parameters
        ----------
        simulation_results : SimulationResults
            Simulation results containing chromatograms.
        components : list, optional
            List of components to be fractionated. Default is None.
        use_total_concentration_components : bool, optional
            Use total concentration components. Default is True.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.
        """
        self.components: Optional[list[str]] = components
        self.use_total_concentration_components: bool = use_total_concentration_components
        self.simulation_results = simulation_results

        super().__init__(*args, **kwargs)

    @property
    def simulation_results(self) -> SimulationResults:
        """SimulationResults: The simulation results containing the chromatograms."""
        return self._simulation_results

    @simulation_results.setter
    def simulation_results(self, simulation_results: SimulationResults) -> None:
        """
        Set the simulation results.

        Parameters
        ----------
        simulation_results : SimulationResults
            Simulation results containing chromatograms.

        Raises
        ------
        TypeError
            If simulation_results is not of type SimulationResults.
        CADETProcessError
            If the simulation results do not contain any chromatograms.
        """
        if not isinstance(simulation_results, SimulationResults):
            raise TypeError("Expected SimulationResults")

        if len(simulation_results.chromatograms) == 0:
            raise CADETProcessError("Simulation results do not contain chromatogram")

        self._simulation_results = simulation_results

        self._chromatograms = [
            slice_solution(
                chrom,
                components=self.components,
                use_total_concentration_components=self.use_total_concentration_components,
            )
            for chrom in simulation_results.chromatograms
        ]

        m_feed = np.zeros((self.component_system.n_comp,))
        counter = 0
        for comp, indices in simulation_results.component_system.indices.items():
            if comp in self.component_system.names:
                m_feed_comp = simulation_results.process.m_feed[indices]
                if self.use_total_concentration_components:
                    m_feed[counter] = np.sum(m_feed_comp)
                    counter += 1
                else:
                    n_species = len(indices)
                    m_feed[counter : counter + n_species] = m_feed_comp
                    counter += n_species
        self.m_feed: np.ndarray = m_feed

        self._fractionation_states = Dict({chrom: [] for chrom in self.chromatograms})
        self._chromatogram_events = Dict({chrom: [] for chrom in self.chromatograms})

        self._cycle_time = self.process.cycle_time

        self.reset()

    @property
    def component_system(self) -> ComponentSystem:
        """ComponentSystem: The component system of the chromatograms."""
        return self.chromatograms[0].component_system

        """Enable calling functions with chromatogram object or name."""

    def _call_by_chrom_name(func: Callable) -> Callable:
        @wraps(func)
        def wrapper_call_by_chrom_name(
            self: "Fractionator", chrom: str | SolutionIO, *args: Any, **kwargs: Any
        ) -> Any:
            """Enable calling functions with chromatogram object or name."""
            if isinstance(chrom, str):
                try:
                    chrom = self.chromatograms_dict[chrom]
                except KeyError:
                    raise CADETProcessError("Not a valid unit")
            return func(self, chrom, *args, **kwargs)

        return wrapper_call_by_chrom_name

    @property
    def chromatograms(self) -> list[SolutionIO]:
        """
        list[SolutionIO]: Chromatograms to be fractionized.

        See Also
        --------
        SolutionIO
        reset
        cycle_time
        """
        return self._chromatograms

    @property
    def chromatograms_dict(self) -> dict[str, SolutionIO]:
        """dict: Chromatogram names and objects."""
        return {chrom.name: chrom for chrom in self.chromatograms}

    @property
    def chromatogram_names(self) -> list[str]:
        """list[str]: Names of chromatogram."""
        return [chrom.name for chrom in self.chromatograms]

    @property
    def n_chromatograms(self) -> int:
        """int: Number of Chromatograms Fractionator."""
        return len(self.chromatograms)

    @property
    def chromatogram_events(self) -> dict[SolutionIO, list[Event]]:
        """dict[SolutionIO, list[Event]]: Events sorted by chromatogram."""
        chrom_events = {
            chrom: sorted(events, key=lambda evt: evt.time)
            for chrom, events in self._chromatogram_events.items()
        }

        return chrom_events

    @property
    def process(self) -> Process:
        """Process: The process from the simulation results."""
        return self.simulation_results.process

    @property
    def n_comp(self) -> int:
        """int: Number of components to be fractionized."""
        return self.chromatograms[0].n_comp

    @property
    def cycle_time(self) -> float:
        """
        The cycle time of the Fractionator.

        Note that in some situations, it might be desired to set a custom cycle time
        for calculating the performance indicators. For this purpose, overwrite the
        cycle time in the Process object after adding it to the Fractionator.

        Warning: This is not a robust feature! Side effects can ocurr in the Process!

        See Also
        --------
        productivity

        """
        return self._cycle_time

    @property
    def time(self) -> np.ndarray:
        """np.ndarray: solution times of Chromatogram."""
        return self.chromatograms[0].time

    @plotting.create_and_save_figure
    def plot_fraction_signal(
        self,
        chromatogram: Optional[SolutionIO | str] = None,
        x_axis_in_minutes: bool = True,
        ax: Optional[Axes] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the signal without the waste fractions.

        Parameters
        ----------
        chromatogram : Optional[SolutionIO | str]
            Chromatogram to be plotted. If None, the first one is plotted.
        ax : Axes, optional
            Axes to plot on. If None, a new figure is created.
        x_axis_in_minutes: bool, optional
            Option to use x-aches (time) in minutes, default is set to True.
        *args : Any
            Optional Parameter passed down to plot function.
        **kwargs : Any
            Additional Parameter passed down to plot function.

        Returns
        -------
        ax : Axes
            Axes with plot of parameter state.

        See Also
        --------
        CADETProcess.plot
        plot_purity
        """
        if chromatogram is None:
            chromatogram = list(
                self.performer_timelines["fractionation_states"].keys()
            )[0]
        if isinstance(chromatogram, str):
            chromatogram = self.chromatograms_dict[chromatogram]

        time_line = self.performer_timelines["fractionation_states"][chromatogram.name]

        try:
            start: float = kwargs["start"]
            if x_axis_in_minutes:
                start = start / 60
        except KeyError:
            start = 0
        try:
            end: float = kwargs["end"]
            if x_axis_in_minutes:
                end = end / 60
        except KeyError:
            end = np.max(chromatogram.time)

        _, ax = chromatogram.plot(
            show=False, ax=ax, x_axis_in_minutes=x_axis_in_minutes, *args, **kwargs
        )

        y_max = 1.1 * np.max(chromatogram.solution)

        fill_regions = []
        for sec in time_line.sections:
            comp_index = int(np.where(sec.coeffs)[0].squeeze())
            if comp_index == self.n_comp:
                color_index = -1
                text = "W"
            else:
                color_index = comp_index
                text = self.component_system.names[comp_index]

            sec_start = sec.start
            sec_end = sec.end

            if x_axis_in_minutes:
                sec_start = sec_start / 60
                sec_end = sec_end / 60

            if sec_start != sec_end:
                fill_regions.append(
                    plotting.FillRegion(
                        start=sec_start,
                        end=sec_end,
                        y_max=y_max,
                        color_index=color_index,
                        text=text,
                    )
                )

        if len(time_line.sections) == 0:
            fill_regions.append(
                plotting.FillRegion(
                    start=sec_start, end=sec_end, y_max=y_max, color_index=-1, text="W"
                )
            )

        plotting.add_fill_regions(ax, fill_regions, (start, end))

        return ax

    @property
    def fractionation_states(self) -> Dict:
        """dict: Fractionation state of Chromatograms.

        Notes
        -----
        This is just a dummy variable to support interfacing with Events.

        """
        return self._fractionation_states

    @_call_by_chrom_name
    def set_fractionation_state(
        self, chrom: SolutionIO, state: int | list[float]
    ) -> None:
        """
        Set fractionation states of Chromatogram.

        Parameters
        ----------
        chrom : SolutionIO
            Chromatogram object which is to be fractionated.
        state : int or list of floats
            New fractionation state of the Chromatogram.

        Raises
        ------
        CADETProcessError
            If Chromatogram not in Fractionator
            If state is integer and the state >= the n_comp.
            If the length of the states is unequal the state_length.
            If the sum of the states is not equal to 1.
        """
        if chrom not in self.chromatograms:
            raise CADETProcessError("Chromatogram not in Fractionator")

        state_length = self.n_comp + 1

        if state_length == 0:
            fractionation_state = []

        if type(state) is int:
            if state >= state_length:
                raise CADETProcessError("Index exceeds fractionation states")

            fractionation_state = [0] * state_length
            fractionation_state[state] = 1
        else:
            if len(state) != state_length:
                raise CADETProcessError(f"Expected length {state_length}.")

            elif sum(state) != 1:
                raise CADETProcessError("Sum of fractions must be 1")

            fractionation_state = state

        self._fractionation_states[chrom] = fractionation_state

    @property
    def n_fractions_per_pool(self) -> list[int]:
        """list: number of fractions per pool."""
        return [pool.n_fractions for pool in self.fraction_pools]

    @property
    def fraction_pools(self) -> list[FractionPool]:
        """
        List of the component and waste fraction pools.

        For every event, the end time is determined and a Fraction object is
        created which holds information about start and end time, as well as
        the mass and the volume of the fraction. The fractions are pooled
        depending on the event state.

        Returns
        -------
        fraction_pools : list
            List with fraction pools.
        """
        if self._fraction_pools is None:
            self._fraction_pools = [
                FractionPool(self.n_comp) for _ in range(self.n_comp + 1)
            ]

            for chrom_index, chrom in enumerate(self.chromatograms):
                chrom_events = self.chromatogram_events[chrom]
                for evt_index, evt in enumerate(chrom_events):
                    target = np.nonzero(evt.full_state)[0].squeeze()

                    frac_start = evt.time

                    if evt_index < len(chrom_events) - 1:
                        frac_end = chrom_events[evt_index + 1].time
                        fraction = self._create_fraction(
                            chrom_index, frac_start, frac_end
                        )
                        self.add_fraction(fraction, target)
                    else:
                        frac_end = self.cycle_time
                        fraction = self._create_fraction(
                            chrom_index, frac_start, frac_end
                        )
                        self.add_fraction(fraction, target)

                        frac_start = 0
                        frac_end = chrom_events[0].time
                        fraction = self._create_fraction(
                            chrom_index, frac_start, frac_end
                        )
                        self.add_fraction(fraction, target)

        return self._fraction_pools

    def _create_fraction(self, chrom_index: int, start: float, end: float) -> Fraction:
        """
        Create Fraction object calculate mass.

        Parameters
        ----------
        chrom_index : int
            index of the chromatogram
        start : float
            start time of the fraction
        end : float
            end time of the fraction

        Returns
        -------
        fraction : Fraction
            Chromatogram fraction
        """
        fraction = self.chromatograms[chrom_index].create_fraction(start, end)

        return fraction

    def add_fraction(self, fraction: Fraction, target: int) -> None:
        """
        Add Fraction to the FractionPool of target component.

        Notes
        -----
        Waste is always the last fraction
        """
        if not isinstance(fraction, Fraction):
            raise TypeError("Expected Fraction")

        if target not in range(self.n_comp + 1):
            raise CADETProcessError("Not a valid target")
        self._fraction_pools[target].add_fraction(fraction)

    @property
    def mass(self) -> np.ndarray:
        """ndarray: Component mass in corresponding fraction pools."""
        if self._mass is None:
            self._mass = np.array([
                pool.mass[comp]
                for comp, pool in enumerate(self.fraction_pools[:-1])
            ])
        return self._mass

    @property
    def total_mass(self) -> np.ndarray:
        """ndarray: Total mass of each component in all fraction pools."""
        return np.sum([pool.mass for pool in self.fraction_pools], axis=0)

    @property
    def concentration(self) -> np.ndarray:
        """ndarray: Component concentration in corresponding fraction pool."""
        return np.array([
            pool.concentration[comp]
            for comp, pool in enumerate(self.fraction_pools[:-1])
        ])

    @property
    def purity(self) -> np.ndarray:
        """ndarray: Component purity in corresponding fraction pool."""
        return np.array([
            pool.purity[comp]
            for comp, pool in enumerate(self.fraction_pools[:-1])
        ])

    @property
    def recovery(self) -> np.ndarray:
        """ndarray: Component recovery yield in corresponding fraction pool."""
        with np.errstate(divide="ignore", invalid="ignore"):
            recovery = self.mass / self.m_feed

        return np.nan_to_num(recovery)

    @property
    def mass_balance_difference(self) -> np.ndarray:
        """
        ndarray: Difference in mass balance between m_feed and fraction pools.

        The mass balance is calculated as the difference between the feed mass (m_feed)
        and the mass in the fraction pools. It represents the discrepancy or change in
        mass during the fractionation process.

        Returns
        -------
        ndarray
            Difference in mass balance between m_feed and fraction pools for each
            component.

        Notes
        -----
        Positive values indicate a surplus of mass in the fraction pools compared to
        the feed, while negative values indicate a deficit. A value of zero indicates
        a mass balance where the mass in the fraction pools is equal to the feed mass.

        """
        return self.total_mass - self.m_feed

    @property
    def productivity(self) -> np.ndarray:
        """ndarray: Specific productivity in corresponding fraction pool."""
        return self.mass / (self.process.cycle_time * self.process.V_solid)

    @property
    def eluent_consumption(self) -> np.ndarray:
        """
        ndarray: Specific eluent consumption in corresponding fraction pool.

        Notes
        -----
        This is the inverse of the regularly used specific eluent
        consumption. It is preferred here in order to avoid numeric issues
        if the collected mass is 0.
        """
        return self.mass / self.process.V_eluent

    @property
    def performance(self) -> Performance:
        """Performance: The performance metrics of the fractionation."""
        self.reset()
        return Performance(
            self.mass,
            self.concentration,
            self.purity,
            self.recovery,
            self.productivity,
            self.eluent_consumption,
            self.mass_balance_difference,
            self.component_system,
        )

    def reset(self) -> None:
        """Reset the results when fractionation times are changed."""
        self._fractionation_state = None
        self._fraction_pools = None
        self._mass = None

    def add_fractionation_event(
        self,
        event_name: str,
        target: str | int,
        time: float,
        chromatogram: Optional[SolutionIO | str] = None,
    ) -> None:
        """
        Add a fractionation event.

        Parameters
        ----------
        event_name : str
            The name of the event.
        target : str | int
            The indice or name of target component in Component System.
        time : float
            The time of the event.
        chromatogram : Optional[SolutionIO | str]
            The chromatogram associated with the event.
            If None and there is only one chromatogram, it will be used.

        Raises
        ------
        CADETProcessError
            If the chromatogram is not found.
        """
        if chromatogram is None and self.n_chromatograms > 1:
            raise CADETProcessError(
                "Missing chromatogram for which the fractionation is added."
            )
        elif chromatogram is None and self.n_chromatograms == 1:
            chromatogram = self.chromatograms[0]

        if isinstance(chromatogram, str):
            try:
                chromatogram = self.chromatograms_dict[f"{chromatogram}"]
            except KeyError:
                raise CADETProcessError("Could not find chromatogram.")

        if chromatogram not in self.chromatograms:
            raise CADETProcessError("Could not find chromatogram.")

        param_path = f"fractionation_states.{chromatogram.name}"

        if isinstance(target, str):
            target = self.component_system.names.index(target)

        evt = self.add_event(event_name, param_path, target, time)
        self._chromatogram_events[chromatogram].append(evt)

        self.reset()

    def initial_values(self, purity_required: float | list[float] = 0.95) -> None:
        """
        Create events from chromatogram with minimum purity.

        Function creates fractions for areas in the chromatogram, where the
        local purity profile is higher than the purity required.

        Parameters
        ----------
        purity_required : float or list of floats
            Minimum purity required for the components in the fractionation

        Raises
        ------
        ValueError
            If the size of the purity parameter does not match the number of components
        """
        if isinstance(purity_required, float):
            purity_required = [purity_required] * self.n_comp
        elif len(purity_required) != self.n_comp:
            raise ValueError(f"Expected purity array with size {self.n_comp}")

        self._events = []
        self._chromatogram_events = Dict({chrom: [] for chrom in self.chromatograms})
        self.reset()

        for chrom_index, chrom in enumerate(self.chromatograms):
            purity_min = np.zeros(chrom.solution.shape)
            purity_min[chrom.local_purity_components > purity_required] = 1
            diff = np.vstack((
                purity_min[0, :] - purity_min[-1, :],
                np.diff(purity_min, axis=0)
            ))

            for comp in range(self.n_comp):
                if purity_required[comp] > 0:
                    on_indices = np.where(diff[:, comp] == 1)[0]
                    off_indices = np.where(diff[:, comp] == -1)[0] - 1

                    # Handle the case where the entire array is above the threshold
                    if (
                        len(on_indices) == 0
                        and len(off_indices) == 0
                        and purity_min[0, comp] == 1
                    ):
                        on_indices = np.array([0])
                        off_indices = np.array([purity_min.shape[0] - 1])

                    # Ensure regions with a single value are ignored
                    valid_regions = [
                        (on, off)
                        for on, off in zip(on_indices, off_indices)
                        if off != on
                    ]

                    for index, (on_evt, off_evt) in enumerate(valid_regions):
                        # Add start event
                        time = chrom.time[int(on_evt)]
                        event_name = f"chrom_{chrom_index}_comp_{comp}_start_{index}"
                        param_path = f"fractionation_states.{chrom.name}"
                        evt = self.add_event(event_name, param_path, comp, time)
                        self._chromatogram_events[chrom].append(evt)

                        # Add end event
                        time = chrom.time[int(off_evt)]
                        event_name = f"chrom_{chrom_index}_comp_{comp}_end_{index}"
                        param_path = f"fractionation_states.{chrom.name}"
                        evt = self.add_event(event_name, param_path, self.n_comp, time)
                        self._chromatogram_events[chrom].append(evt)

        if not self.check_duplicate_events():
            chrom_events = self._chromatogram_events.copy()
            for chrom, events in chrom_events.items():
                events_at_time = defaultdict(list)
                for event in events:
                    events_at_time[event.time].append(event)

                for events in events_at_time.values():
                    if len(events) == 1:
                        continue
                    for evt in events:
                        if evt.state == self.n_comp:
                            self.remove_event(evt.name)
                            self._chromatogram_events[chrom].remove(evt)

    @property
    def parameters(self) -> Dict:
        """dict: Parameters of the fractionator."""
        parameters = super().parameters
        parameters["fractionation_states"] = {
            chrom.name: self.fractionation_states[chrom] for chrom in self.chromatograms
        }

        return Dict(parameters)

    @parameters.setter
    def parameters(self, parameters: Dict) -> None:
        try:
            fractionation_states = parameters.pop("fractionation_states")
            for chrom, state in fractionation_states.items():
                self.set_fractionation_state(chrom, state)
        except KeyError:
            pass

        super(Fractionator, self.__class__).parameters.fset(self, parameters)

    @property
    def section_dependent_parameters(self) -> Dict:
        """dict: Section dependent parameters of the fractionator."""
        return self.parameters

    def save(
        self,
        case_dir: str,
        start: float = 0,
        end: Optional[float] = None,
    ) -> None:
        """
        Save chromatogram and purity plots to a specified directory.

        Parameters
        ----------
        case_dir : str
            Directory name within the working directory to save plots.
        start : float, optional
            Start time for plotting purity, default is 0.
        end : Optional[float]
            End time for plotting purity. If None, includes all data.
        """
        path = os.path.join(settings.working_directory, case_dir)

        for index, chrom in enumerate(self.chromatograms):
            chrom.plot(save_path=path + f"/chrom_{index}.png")
            chrom.plot_purity(
                start=start, end=end, save_path=path + "/chrom_purity.png"
            )

        for chrom in enumerate(self.chromatograms):
            self.plot_fraction_signal(
                chromatogram=chrom,
                start=start,
                end=end,
                save_path=path + f"/fractionation_signal_{index}.png",
                index=index,
            )

    def __str__(self) -> str:
        """str: String representation of the fractionator."""
        return self.__class__.__name__



from typing import Any, Optional

from .dataStructure import Structure
from .parameter import Bool


class cached_property_if_locked(property):
    """
    A property that caches its value if the instance is locked.

    This property extends the built-in property to cache its value when the instance
    is locked. The cached value is stored in the instance's `cached_properties` dictionary.
    """

    def __get__(self, instance: Any, cls: Optional[type] = None) -> Any:
        """
        Get the value of the property, using the cache if the instance is locked.

        Parameters
        ----------
        instance : Any
            The instance from which to retrieve the property value.
        cls : Optional[type], optional
            The class of the instance, by default None.

        Returns
        -------
        Any
            The value of the property.
        """
        if instance.lock:
            try:
                return instance.cached_properties[self.name]
            except KeyError:
                pass

        value = super().__get__(instance, cls)

        if instance.lock:
            instance.cached_properties[self.name] = value

        return value

    @property
    def name(self) -> str:
        """str: name of the property."""
        return self.fget.__name__


class CachedPropertiesMixin(Structure):
    """
    Mixin class for caching properties in a structured object.

    This class is designed to be used as a mixin in conjunction with other classes
    inheriting from `Structure`. It provides functionality for caching properties and
    managing a lock state to control the caching behavior.

    Notes
    -----
    - To prevent the return of outdated state, the cache is cleared whenever the `lock`
      state is changed.
    """

    _lock = Bool(default=False)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize Cached Properties Mixin Object."""
        super().__init__(*args, **kwargs)
        self.cached_properties = {}

    @property
    def lock(self) -> bool:
        """bool: If True, properties are cached. False otherwise."""
        return self._lock

    @lock.setter
    def lock(self, lock: bool) -> None:
        self._lock = lock
        self.cached_properties = {}



import unittest

import numpy as np
from CADETProcess import CADETProcessError
from CADETProcess.processModel import ComponentSystem
from CADETProcess.reference import ReferenceIO
from CADETProcess.solution import SolutionIO
from scipy import stats

comp_2 = ComponentSystem(["A", "B"])

time = np.linspace(0, 100, 1001)

mu_0 = 30
sigma_0 = 5

mu_1 = 40
sigma_1 = 5

mu_2 = 50
sigma_2 = 3


solution_2_gaussian = np.zeros((len(time), 2))
solution_2_gaussian[:, 0] = stats.norm.pdf(time, mu_0, sigma_0)
solution_2_gaussian[:, 1] = stats.norm.pdf(time, mu_1, sigma_1)

solution_2_gaussian_switched = np.zeros((len(time), 2))
solution_2_gaussian_switched[:, 1] = stats.norm.pdf(time, mu_0, sigma_0)
solution_2_gaussian_switched[:, 0] = stats.norm.pdf(time, mu_1, sigma_1)

solution_2_gaussian_different_height = np.zeros((len(time), 2))
solution_2_gaussian_different_height[:, 1] = stats.norm.pdf(time, mu_0, sigma_0)
solution_2_gaussian_different_height[:, 0] = stats.norm.pdf(time, mu_2, sigma_2)

q_const = np.ones(time.shape)


from CADETProcess.comparison import SSE


class TestSSE(unittest.TestCase):
    def test_metric(self):
        # Compare with itself
        component_system = ComponentSystem(1)
        reference = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian[:, [0]],
            component_system=component_system,
        )

        difference = SSE(reference)
        metrics_expected = [0]
        metrics = difference.evaluate(reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gaussian Peak
        component_system = ComponentSystem(1)
        solution = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian[:, [1]],
            component_system=component_system,
        )

        difference = SSE(reference, resample=False)
        metrics_expected = [0.7132717]
        metrics = difference.evaluate(solution)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        difference = SSE(reference, resample=True)
        metrics_expected = [3.56635829]
        metrics = difference.evaluate(solution)
        np.testing.assert_almost_equal(metrics, metrics_expected)


from CADETProcess.comparison import NRMSE


class TestNRMSE(unittest.TestCase):
    def test_metric(self):
        # Compare with itself
        component_system = ComponentSystem(1)
        reference = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian[:, [0]],
            component_system=component_system,
        )

        difference = NRMSE(reference)
        metrics_expected = [0]
        metrics = difference.evaluate(reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gaussian Peak
        component_system = ComponentSystem(1)
        solution = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian[:, [1]],
            component_system=component_system,
        )

        difference = NRMSE(reference, resample=False)
        metrics_expected = [0.3345572]
        metrics = difference.evaluate(solution)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        difference = NRMSE(reference, resample=True)
        metrics_expected = [0.33469097]
        metrics = difference.evaluate(solution)
        np.testing.assert_almost_equal(metrics, metrics_expected)


from CADETProcess.comparison import PeakHeight


class TestPeakHeight(unittest.TestCase):
    def setUp(self):
        # 2 Components, gaussian peaks, constant flow
        component_system = ComponentSystem(1)
        self.reference_single = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian[:, [0]],
            component_system=component_system,
        )

        self.reference = ReferenceIO(
            "simple", time, solution_2_gaussian, component_system=comp_2
        )

        self.reference_switched = ReferenceIO(
            "simple", time, solution_2_gaussian_switched, component_system=comp_2
        )

        self.reference_different_height = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian_different_height,
            component_system=comp_2,
        )

    def test_metric(self):
        # Compare with itself
        difference = PeakHeight(self.reference, components=["A"])
        metrics_expected = [0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak
        difference = PeakHeight(
            self.reference_switched, components=["A"], normalize_metrics=False
        )
        metrics_expected = [0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak, normalize_metrics
        difference = PeakHeight(
            self.reference_switched, components=["A"], normalize_metrics=True
        )
        metrics_expected = [0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Multi-component, compare with self.
        difference = PeakHeight(self.reference)
        metrics_expected = [0, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Swich components
        difference = PeakHeight(self.reference, normalize_metrics=False)
        metrics_expected = [0, 0]
        metrics = difference.evaluate(self.reference_switched)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Different Peaks
        difference = PeakHeight(self.reference, normalize_metrics=False)
        metrics_expected = [0.0531923, 0.0]
        metrics = difference.evaluate(self.reference_different_height)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        difference = PeakHeight(self.reference, normalize_metrics=True)
        metrics_expected = [0.3215127, 0.0]
        metrics = difference.evaluate(self.reference_different_height)
        np.testing.assert_almost_equal(metrics, metrics_expected)


from CADETProcess.comparison import PeakPosition


class TestPeakPosition(unittest.TestCase):
    def setUp(self):
        # 2 Components, gaussian peaks, constant flow
        component_system = ComponentSystem(1)
        self.reference_single = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian[:, [0]],
            component_system=component_system,
        )

        self.reference = ReferenceIO(
            "simple", time, solution_2_gaussian, component_system=comp_2
        )

        self.reference_switched = ReferenceIO(
            "simple", time, solution_2_gaussian_switched, component_system=comp_2
        )

    def test_metric(self):
        # Compare with itself
        difference = PeakPosition(self.reference, components=["A"])
        metrics_expected = [0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak
        difference = PeakPosition(
            self.reference_switched, components=["A"], normalize_metrics=False
        )
        metrics_expected = [10]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak, normalize_metrics
        difference = PeakPosition(
            self.reference_switched, components=["B"], normalize_metrics=True
        )
        metrics_expected = [0.1651404]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Multi-component, compare with self.
        difference = PeakPosition(self.reference)
        metrics_expected = [0, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Swich components
        difference = PeakPosition(self.reference, normalize_metrics=False)
        metrics_expected = [10, 10]
        metrics = difference.evaluate(self.reference_switched)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        difference = PeakPosition(self.reference)
        metrics_expected = [0.1651404, 0.124353]
        metrics = difference.evaluate(self.reference_switched)
        np.testing.assert_almost_equal(metrics, metrics_expected)


from CADETProcess.comparison import Shape


class TestShape(unittest.TestCase):
    def setUp(self):
        # 2 Components, gaussian peaks, constant flow
        component_system = ComponentSystem(1)
        self.reference_single = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian[:, [0]],
            component_system=component_system,
        )

        self.reference = ReferenceIO(
            "simple", time, solution_2_gaussian, component_system=comp_2
        )

        self.reference_switched = ReferenceIO(
            "simple", time, solution_2_gaussian_switched, component_system=comp_2
        )

    def test_metric(self):
        # Compare with itself
        difference = Shape(
            self.reference,
            use_derivative=False,
            components=["A"],
            normalize_metrics=False,
        )
        metrics_expected = [0, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak
        difference = Shape(
            self.reference_switched,
            use_derivative=False,
            components=["A"],
            normalize_metrics=False,
        )
        metrics_expected = [5.5511151e-16, 10]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak, normalize_metrics
        difference = Shape(
            self.reference_switched,
            use_derivative=False,
            components=["A"],
            normalize_metrics=True,
        )
        metrics_expected = [0, 4.6211716e-01]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak, include derivative
        difference = Shape(
            self.reference_switched,
            use_derivative=True,
            components=["A"],
            normalize_metrics=False,
        )
        metrics_expected = [0, 10, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak, include derivative, normalize metrics
        difference = Shape(
            self.reference_switched,
            use_derivative=True,
            components=["A"],
            normalize_metrics=True,
        )
        metrics_expected = [0, 4.6211716e-01, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Multi-component, currently not implemented
        with self.assertRaises(CADETProcessError):
            difference = Shape(self.reference, use_derivative=False)
            metrics_expected = [0, 0]
        with self.assertRaises(CADETProcessError):
            difference = Shape(self.reference_single)
            metrics = difference.evaluate(self.reference)


from CADETProcess.comparison import ShapeFront


class TestShapeFront(unittest.TestCase):
    def setUp(self):
        # 2 Components, gaussian peaks, constant flow
        component_system = ComponentSystem(1)
        self.reference_single = ReferenceIO(
            "simple",
            time,
            solution_2_gaussian[:, [0]],
            component_system=component_system,
        )

        self.reference = ReferenceIO(
            "simple", time, solution_2_gaussian, component_system=comp_2
        )

        self.reference_switched = ReferenceIO(
            "simple", time, solution_2_gaussian_switched, component_system=comp_2
        )

    def test_metric(self):
        # Compare with itself
        difference = ShapeFront(
            self.reference,
            use_derivative=False,
            components=["A"],
            normalize_metrics=False,
        )
        metrics_expected = [0, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak
        difference = ShapeFront(
            self.reference_switched,
            use_derivative=False,
            components=["A"],
            normalize_metrics=False,
        )
        metrics_expected = [0, 10]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak, normalize_metrics
        difference = ShapeFront(
            self.reference_switched,
            use_derivative=False,
            components=["A"],
            normalize_metrics=True,
        )
        metrics_expected = [0, 4.6211716e-01]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak, include derivative
        difference = ShapeFront(
            self.reference_switched,
            use_derivative=True,
            components=["A"],
            normalize_metrics=False,
        )
        metrics_expected = [0, 10, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Compare with other Gauss Peak, include derivative, normalize metrics
        difference = ShapeFront(
            self.reference_switched,
            use_derivative=True,
            components=["A"],
            normalize_metrics=True,
        )
        metrics_expected = [0, 4.6211716e-01, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Use maximum slope
        difference = ShapeFront(
            self.reference,
            use_derivative=False,
            use_max_slope=True,
            components=["A"],
            normalize_metrics=False,
        )
        metrics_expected = [0, 0]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        difference = ShapeFront(
            self.reference_switched,
            use_derivative=False,
            use_max_slope=True,
            components=["A"],
            normalize_metrics=False,
        )
        metrics_expected = [0, 10]
        metrics = difference.evaluate(self.reference)
        np.testing.assert_almost_equal(metrics, metrics_expected)

        # Multi-component, currently not implemented
        with self.assertRaises(CADETProcessError):
            difference = ShapeFront(self.reference, use_derivative=False)
            metrics_expected = [0, 0]
        with self.assertRaises(CADETProcessError):
            difference = ShapeFront(self.reference_single)
            metrics = difference.evaluate(self.reference)


from CADETProcess.comparison import FractionationSSE
from CADETProcess.fractionation import Fraction
from CADETProcess.reference import FractionationReference


class TestFractionation(unittest.TestCase):
    def setUp(self):
        fraction_1 = Fraction(
            start=15,
            end=30,
            mass=[0.49865015, 0.02274985],
            volume=15,
        )
        fraction_2 = Fraction(
            start=30,
            end=45,
            mass=[0.49865015, 0.81859462],
            volume=15,
        )
        self.fractions = [fraction_1, fraction_2]

        component_system = ComponentSystem(["A", "B"])
        self.reference = FractionationReference(
            "fractions", [fraction_1, fraction_2], component_system=component_system
        )

        self.solution = SolutionIO(
            "simple", comp_2, time, solution_2_gaussian, flow_rate=q_const
        )

    def test_metric(self):
        # Compare with itself
        difference = FractionationSSE(
            self.reference,
            components=["A"],
        )
        metrics_expected = [1.30315857e-19]
        metrics = difference.evaluate(self.solution)
        np.testing.assert_almost_equal(metrics, metrics_expected)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from CADETProcess import equilibria
from CADETProcess.processModel import (
    ComponentSystem,
    Langmuir,
    Linear,
    MassActionLaw,
    StericMassAction,
)


class TestReactionEquilibrium(unittest.TestCase):
    def setUp(self):
        component_system = ComponentSystem(2)

        self.single_0 = MassActionLaw(component_system, name="simple")
        self.single_0.add_reaction([0, 1], [-1, 1], 2, k_bwd=2)

        self.single_1 = MassActionLaw(component_system, name="simple")
        self.single_1.add_reaction([0, 1], [-1, 1], 1, k_bwd=1)

        component_system = ComponentSystem(3)
        self.single_2 = MassActionLaw(component_system, name="simple")
        self.single_2.add_reaction([0, 1], [-1, 1], 1, k_bwd=1)

        self.multi = MassActionLaw(component_system, name="simple")
        self.multi.add_reaction([0, 1], [-1, 1], 1, k_bwd=1)
        self.multi.add_reaction([1, 2], [-1, 1], 1, k_bwd=1)

        component_system_lysine = ComponentSystem()
        component_system_lysine.add_component(
            "Lysine",
            species=["Lys2+", "Lys+", "Lys", "Lys-"],
        )
        component_system_lysine.add_component(
            "H+",
        )
        self.lysine = MassActionLaw(component_system_lysine, name="Lysine")
        self.lysine.add_reaction(
            [0, 1, -1], [-1, 1, 1], 10 ** (-2.20) * 1e3, is_kinetic=False
        )
        self.lysine.add_reaction(
            [1, 2, -1], [-1, 1, 1], 10 ** (-8.90) * 1e3, is_kinetic=False
        )
        self.lysine.add_reaction(
            [2, 3, -1], [-1, 1, 1], 10 ** (-10.28) * 1e3, is_kinetic=False
        )

    def test_dydx(self):
        buffer = [1, 0]
        dydx = equilibria.dydx_mal(buffer, self.single_1)
        dydx_expected = [-1, 1]
        np.testing.assert_almost_equal(dydx, dydx_expected)

        buffer = [1, 0, 0]
        dydx = equilibria.dydx_mal(buffer, self.single_2)
        dydx_expected = [-1, 1, 0]
        np.testing.assert_almost_equal(dydx, dydx_expected)

        buffer = [1, 0, 0]
        dydx = equilibria.dydx_mal(buffer, self.single_2, constant_indices=[0])
        dydx_expected = [0, 1, 0]
        np.testing.assert_almost_equal(dydx, dydx_expected)

        buffer = [1, 0, 0]
        buffer_init = [2, 0, 0]
        dydx = equilibria.dydx_mal(
            buffer, self.single_2, constant_indices=[0], c_init=buffer_init
        )
        dydx_expected = [0, 2, 0]
        np.testing.assert_almost_equal(dydx, dydx_expected)

        buffer = [1, 0, 0]
        dydx = equilibria.dydx_mal(buffer, self.multi)
        dydx_expected = [-1, 1, 0]
        np.testing.assert_almost_equal(dydx, dydx_expected)

        buffer = [1, 1, 0]
        dydx = equilibria.dydx_mal(buffer, self.multi)
        dydx_expected = [0, -1, 1]
        np.testing.assert_almost_equal(dydx, dydx_expected)

        buffer = [1, 1, 1]
        dydx = equilibria.dydx_mal(buffer, self.multi)
        dydx_expected = [0, 0, 0]
        np.testing.assert_almost_equal(dydx, dydx_expected)

        buffer = [0, 1, 0]
        dydx = equilibria.dydx_mal(buffer, self.multi)
        dydx_expected = [1, -2, 1]
        np.testing.assert_almost_equal(dydx, dydx_expected)

    def test_jac(self):
        buffer = [1, 0]
        jac = equilibria.jac_mal(buffer, self.single_0)
        jac_expected = [[-2, 2], [2, -2]]
        np.testing.assert_almost_equal(jac, jac_expected)

    def test_reaction_equilibrium(self):
        buffer = [1, 0]

        eq = equilibria.calculate_buffer_equilibrium(buffer, self.single_1)
        eq_expected = [0.5, 0.5]
        np.testing.assert_almost_equal(eq, eq_expected)

        buffer = [0, 0, 0, 1, 0]

        pH = 0
        buffer[-1] = 10 ** (-pH + 3)
        eq = equilibria.calculate_buffer_equilibrium(
            buffer, self.lysine, constant_indices=[-1]
        )
        eq_expected = [0.99373000042703, 0.006270012422, 7.89e-12, 0.0, buffer[-1]]
        pH = 3
        buffer[-1] = 10 ** (-pH + 3)
        eq = equilibria.calculate_buffer_equilibrium(
            buffer, self.lysine, constant_indices=[-1]
        )
        eq_expected = [
            0.13680673998909,
            0.86319217370507,
            1.08669456e-06,
            6e-14,
            buffer[-1],
        ]
        np.testing.assert_almost_equal(eq, eq_expected)
        pH = 7
        buffer[-1] = 10 ** (-pH + 3)
        eq = equilibria.calculate_buffer_equilibrium(
            buffer, self.lysine, constant_indices=[-1]
        )
        eq_expected = [
            1.565153925e-05,
            0.98754536426943,
            0.01243245954378,
            6.52464752e-06,
            buffer[-1],
        ]
        np.testing.assert_almost_equal(eq, eq_expected)
        pH = 11
        buffer[-1] = 10 ** (-pH + 3)
        eq = equilibria.calculate_buffer_equilibrium(
            buffer, self.lysine, constant_indices=[-1]
        )
        eq_expected = [
            2.01e-12,
            0.00126970262768,
            0.15984609034134,
            0.83888420702896,
            buffer[-1],
        ]
        np.testing.assert_almost_equal(eq, eq_expected)
        pH = 14
        buffer[-1] = 10 ** (-pH + 3)
        eq = equilibria.calculate_buffer_equilibrium(
            buffer, self.lysine, constant_indices=[-1]
        )
        eq_expected = [0.0, 0.0, 1e-11, 0.99999999999, buffer[-1]]


class TestAdsorptionEquilibrium(unittest.TestCase):
    def setUp(self):
        component_system_mono = ComponentSystem()
        component_system_mono.add_component("A")

        component_system_di = ComponentSystem()
        component_system_di.add_component("A")
        component_system_di.add_component("B")

        self.linear = Linear(component_system_mono, "linear")
        self.linear.adsorption_rate = [1]
        self.linear.desorption_rate = [1]

        self.langmuir = Langmuir(component_system_di, "langmuir")
        self.langmuir.adsorption_rate = [2, 1]
        self.langmuir.desorption_rate = [1, 1]
        self.langmuir.capacity = [10, 10]

        self.sma = StericMassAction(component_system_di, "SMA")
        self.sma.adsorption_rate = [1, 2]
        self.sma.desorption_rate = [1, 1]
        self.sma.characteristic_charge = [1, 1]
        self.sma.steric_factor = [0, 0]
        self.sma.capacity = 10

    def test_adsorption(self):
        buffer = [1]
        eq = equilibria.simulate_solid_equilibria(self.linear, buffer)
        eq_expected = [1]
        np.testing.assert_almost_equal(eq, eq_expected)

        buffer = [1, 1]
        eq = equilibria.simulate_solid_equilibria(self.langmuir, buffer)
        eq_expected = [5, 2.5]
        np.testing.assert_almost_equal(eq, eq_expected)

        buffer = [1, 1]
        eq = equilibria.simulate_solid_equilibria(self.sma, buffer)
        eq_expected = [10 / 3, 2 * 10 / 3]
        np.testing.assert_almost_equal(eq, eq_expected, decimal=4)


if __name__ == "__main__":
    unittest.main()



import copy

import pytest
from CADETProcess.dataStructure.nested_dict import (
    check_nested,
    generate_nested_dict,
    get_leaves,
    get_nested_attribute,
    get_nested_list_value,
    get_nested_value,
    insert_path,
    set_nested_attribute,
    set_nested_list_value,
    set_nested_value,
    update_dict_recursively,
)


@pytest.fixture
def sample_dict():
    """Fixture providing a sample nested dictionary."""
    return {
        "a": {
            "b": {
                "c": 42,
            },
        },
        "x": {
            "y": {
                "z": 99,
            },
        },
    }


@pytest.fixture
def sample_list():
    """Fixture providing a sample nested list."""
    return [[[1, 2], [3, 4]], [[5, 6], [7, 8]]]


# --- TESTS FOR NESTED DICTIONARY FUNCTIONS --- #


def test_check_nested(sample_dict):
    assert check_nested(sample_dict, "a.b.c") is True
    assert check_nested(sample_dict, "a.b") is False
    assert check_nested(sample_dict, "x.y.z") is True
    assert check_nested(sample_dict, "missing.path") is False


def test_generate_nested_dict():
    expected = {"a": {"b": {"c": 10}}}
    assert generate_nested_dict("a.b.c", 10) == expected


def test_insert_path(sample_dict):
    insert_path(sample_dict, "a.b.d", 100)
    assert sample_dict["a"]["b"]["d"] == 100

    insert_path(sample_dict, "new.path.here", 50)
    assert sample_dict["new"]["path"]["here"] == 50


def test_get_leaves(sample_dict):
    leaves = set(get_leaves(sample_dict))
    assert leaves == {"a.b.c", "x.y.z"}


def test_set_nested_value(sample_dict):
    set_nested_value(sample_dict, "a.b.c", 77)
    assert sample_dict["a"]["b"]["c"] == 77

    set_nested_value(sample_dict, "x.y.z", "test")
    assert sample_dict["x"]["y"]["z"] == "test"


def test_get_nested_value(sample_dict):
    assert get_nested_value(sample_dict, "a.b.c") == 42
    assert get_nested_value(sample_dict, "x.y.z") == 99

    with pytest.raises(KeyError):
        get_nested_value(sample_dict, "missing.path")


# --- TESTS FOR DICTIONARY UPDATING --- #


def test_update_dict_recursively(sample_dict):
    target = {"a": {"b": 1}, "c": 3}
    updates = {"a": {"b": 2, "d": 4}, "c": 30, "e": 5}

    updated = update_dict_recursively(copy.deepcopy(target), updates)
    assert updated == {"a": {"b": 2, "d": 4}, "c": 30, "e": 5}

    updated_existing = update_dict_recursively(
        copy.deepcopy(target), updates, only_existing_keys=True
    )
    assert updated_existing == {"a": {"b": 2}, "c": 30}


# --- TESTS FOR OBJECT ATTRIBUTE FUNCTIONS --- #


class SampleObject:
    def __init__(self):
        self.a = SampleSubObject()


class SampleSubObject:
    def __init__(self):
        self.b = SampleInnerObject()


class SampleInnerObject:
    def __init__(self):
        self.c = 42


@pytest.fixture
def sample_obj():
    return SampleObject()


def test_get_nested_attribute(sample_obj):
    assert get_nested_attribute(sample_obj, "a.b.c") == 42

    with pytest.raises(AttributeError):
        get_nested_attribute(sample_obj, "a.b.d")


def test_set_nested_attribute(sample_obj):
    set_nested_attribute(sample_obj, "a.b.c", 99)
    assert sample_obj.a.b.c == 99

    with pytest.raises(AttributeError):
        set_nested_attribute(sample_obj, "a.c.b", 50)


# --- TESTS FOR NESTED LIST FUNCTIONS --- #


def test_get_nested_list_value(sample_list):
    assert get_nested_list_value(sample_list, (0, 1, 1)) == 4
    assert get_nested_list_value(sample_list, (1, 0, 1)) == 6

    with pytest.raises(IndexError):
        get_nested_list_value(sample_list, (3, 2, 1))


def test_set_nested_list_value(sample_list):
    set_nested_list_value(sample_list, (0, 1, 1), 99)
    assert sample_list[0][1][1] == 99

    with pytest.raises(IndexError):
        set_nested_list_value(sample_list, (3, 2, 1), 100)


if __name__ == "__main__":
    pytest.main([__file__])



from typing import Any

from CADETProcess import CADETProcessError


class ParameterWrapper:
    """
    Base class for converting the config from objects such as units.

    Attributes
    ----------
    _base_class : type
        Type constraint for wrapped object
    _wrapped_object : obj
        Object whose config is to be converted

    Raises
    ------
    CADETProcessError
        If the wrapped_object is no instance of the base_class.
    """

    _base_class = object

    def __init__(self, wrapped_object: Any) -> None:
        """Construct ParameterWrapper object."""
        if not isinstance(wrapped_object, self._baseClass):
            raise CADETProcessError(f"Expected {self._baseClass}")

        model = wrapped_object.model
        try:
            self.model_parameters = self._model_parameters[model]
        except KeyError:
            raise CADETProcessError("Model Type not defined")

        self._wrapped_object = wrapped_object

    @property
    def parameters(self) -> dict:
        """dict: Parameters dictionary."""
        model_parameters = {}

        model_parameters[self._model_type] = self.model_parameters["name"]

        for key, value in self.model_parameters["parameters"].items():
            value = getattr(self._wrapped_object, value)
            if value is not None:
                model_parameters[key] = value

        for key, value in self.model_parameters.get("fixed", dict()).items():
            if isinstance(value, list):
                value = self._wrapped_object.n_comp * value
            model_parameters[key] = value

        return model_parameters



from functools import wraps
from typing import Optional

import numpy as np
import numpy.typing as npt

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import String, StructMeta, UnsignedInteger
from CADETProcess.processModel import (
    BindingBaseClass,
    BulkReactionBase,
    ComponentSystem,
    Cstr,
    FlowSheet,
    Inlet,
    Outlet,
    ParticleReactionBase,
    Process,
    ReactionBaseClass,
)


class CompartmentBuilder(metaclass=StructMeta):
    """
    Class to build complex compartment models of bioreactors.

    Attributes
    ----------
    name : str
        Name of the CompartmentBuilder.
    n_compartments : int
        Number of compartments in the bioreactor.
    """

    name = String()
    n_compartments = UnsignedInteger()

    def __init__(
        self,
        component_system: ComponentSystem,
        compartment_volumes: list[float],
        flow_rate_matrix: list[float],
        init_c: int | float | list | np.ndarray = 0,
        binding_model: Optional[BindingBaseClass] = None,
        bulk_reaction_model: Optional[BulkReactionBase] = None,
        particle_reaction_model: Optional[ParticleReactionBase] = None,
        name: Optional[str] = None,
    ) -> None:
        """
        Initialize builder.

        Parameters
        ----------
        component_system : component_system
            DESCRIPTION.
        compartment_volumes : list
            Volume of compartments.
        flow_rate_matrix : list
            Flow rates between compartments.
        init_c : int, float, list, np.array, optional
            Initial concentrations. The default is 0.
        binding_model : BindingBaseClass, optional
            Binding model to set for all compartments. The default is None.
        bulk_reaction_model : TYPE, optional
            Bulk reaction model for all compartments. The default is None.
        particle_reaction_model : TYPE, optional
            Particle reaction model for all compartments. The default is None.
        name : str, optional
            Name of the model. The default is None.
        """
        self.component_system = component_system
        if name is None:
            name = "CompartmentBuilder"
        self.name = name

        self._compartment_model = CompartmentModel(
            component_system, "master compartment"
        )

        self.binding_model = binding_model
        self.bulk_reaction_model = bulk_reaction_model
        self.particle_reaction_model = particle_reaction_model

        self._flow_sheet = FlowSheet(component_system, name)
        self._process = Process(self._flow_sheet, name)

        self._add_compartments(compartment_volumes)
        self._add_connections(flow_rate_matrix)

        self.init_c = init_c

    @property
    def n_comp(self) -> int:
        """int: Number of components."""
        return self.component_system.n_comp

    @property
    def _real_compartments(self) -> list[str]:
        """list: Compartment units excluding pseudo units s.a. Inlet/Outlet."""
        compartments = []
        for i in range(self.n_compartments):
            name = f"compartment_{i}"
            compartment = self.flow_sheet[name]
            if isinstance(compartment, Cstr):
                compartments.append(compartment)

        return compartments

    @property
    @wraps(Cstr.binding_model)
    def binding_model(self) -> ReactionBaseClass:
        """Wrapper around master compartment to set binding model."""
        return self._compartment_model.binding_model

    @binding_model.setter
    @wraps(Cstr.binding_model)
    def binding_model(self, binding_model: BindingBaseClass) -> None:
        if binding_model is not None:
            self._compartment_model.binding_model = binding_model

            for compartment in self._real_compartments:
                compartment.binding_model = binding_model

    @property
    @wraps(Cstr.bulk_reaction_model)
    def bulk_reaction_model(self) -> BindingBaseClass:
        """Wrapper around master compartment to set bulk reaction model."""
        return self._compartment_model.bulk_reaction_model

    @bulk_reaction_model.setter
    @wraps(Cstr.bulk_reaction_model)
    def bulk_reaction_model(self, bulk_reaction_model: BulkReactionBase) -> None:
        if bulk_reaction_model is not None:
            self._compartment_model.bulk_reaction_model = bulk_reaction_model

            for compartment in self._real_compartments:
                compartment.bulk_reaction_model = bulk_reaction_model

    @property
    @wraps(Cstr.particle_reaction_model)
    def particle_reaction_model(self) -> ReactionBaseClass:
        """Wrapper around master compartment to set particle reaction model."""
        return self._compartment_model.particle_reaction_model

    @particle_reaction_model.setter
    @wraps(Cstr.particle_reaction_model)
    def particle_reaction_model(
        self, particle_reaction_model: ParticleReactionBase
    ) -> None:
        if particle_reaction_model is not None:
            self._compartment_model.particle_reaction_model = particle_reaction_model

            for compartment in self._real_compartments:
                compartment.particle_reaction_model = particle_reaction_model

    @property
    def flow_sheet(self) -> FlowSheet:
        """FlowSheet: FlowSheet of the compartment builder."""
        return self._flow_sheet

    @property
    def process(self) -> Process:
        """Process: Process of the compartment builder."""
        return self._process

    @property
    def cycle_time(self) -> float:
        """float: Cycle time of the process."""
        return self.process.cycle_time

    @cycle_time.setter
    def cycle_time(self, cycle_time: float) -> None:
        self.process.cycle_time = cycle_time

    def _add_compartments(self, compartment_volumes: list[float]) -> None:
        """Instantiate compartments and add to FlowSheet."""
        self.n_compartments = len(compartment_volumes)

        for i, vol in enumerate(compartment_volumes):
            name = f"compartment_{i}"

            if vol == "inlet":
                unit = Inlet(self.component_system, name)
            elif vol == "outlet":
                unit = Outlet(self.component_system, name)
            else:
                unit = Cstr(self.component_system, name)
                unit.init_liquid_volume = vol

            self.flow_sheet.add_unit(unit)

    def _add_connections(self, connections_matrix: npt.ArrayLike) -> None:
        """Add connections and flow rates between compartments to FlowSheet."""
        try:
            if isinstance(connections_matrix, list):
                arr = np.array(connections_matrix)
                connections_matrix = arr.reshape(-1, self.n_compartments)
        except ValueError:
            raise CADETProcessError("Expected square matrix")

        if connections_matrix.shape[0] != connections_matrix.shape[1]:
            raise CADETProcessError("Expected square matrix")

        for iOrigin, destinations in enumerate(connections_matrix):
            flow_rate = np.sum(destinations)

            origin = self.flow_sheet[f"compartment_{iOrigin}"]
            output_state = []

            for iDestination, flow in enumerate(destinations):
                if flow == 0:
                    continue

                destination = self.flow_sheet[f"compartment_{iDestination}"]
                output_state.append(flow / flow_rate)

                self.flow_sheet.add_connection(origin, destination)

            if isinstance(origin, Outlet):
                continue

            self.flow_sheet[origin.name].flow_rate = flow_rate
            self.flow_sheet.set_output_state(origin, output_state)

    @property
    def init_c(self) -> int | float | list | np.ndarray:
        """
        np.array: Initial conditions of compartments.

        Parameters
        ----------
        init_c : int, float, list, or np.array
            Initial concentration of compartments.
            If type is float or int, same value is used for all components
            and concentrations.
            If type is list, same component concentrations are used for all
            compartments.
            If np.array, explicit concentrations for components and
            compartments are set.

        Raises
        ------
        ValueError
            If init_c does not contain correct shape.
        """
        return self._init_c

    @init_c.setter
    def init_c(self, init_c: int | float | list | np.ndarray) -> None:
        if isinstance(init_c, (int, float)):
            init_c = init_c * np.ones((self.n_compartments, self.n_comp))
        elif isinstance(init_c, list) and len(init_c) == self.n_comp:
            init_c = np.tile(init_c, (self.n_compartments, 1))
        elif isinstance(init_c, np.ndarray) and init_c.shape == (
            self.n_compartments,
            self.n_comp,
        ):
            pass
        else:
            raise ValueError("unexpected value for init_c")

        self._init_c = init_c

        for i in range(self.n_compartments):
            compartment = self.flow_sheet[f"compartment_{i}"]
            if not isinstance(compartment, Outlet):
                compartment.c = init_c[i, :].tolist()

    def add_tracer(
        self,
        compartment_index: int,
        c: list,
        t_inj: float,
        flow_rate: Optional[float],
        t_start: Optional[float] = 0,
        flow_rate_filter: Optional[bool] = True,
    ) -> None:
        """
        Add tracer injection to compartment model.

        For this purpose, a new inlet source is instantiated and connected to
        the corresponding compartment. Then, an Event is added which modifies
        the flow rate of the inlet.

        Parameters
        ----------
        compartment_index : int
            Compartment to which tracer is injected.
        c : list
            Tracer concentration.
        t_inj : float
            Length of injection.
        flow_rate : float
            flow rate during injection.
        t_start : float, optional
            Time at which injection starts. The default is 0.
        flow_rate_filter : bool, optional
            If True, the compartment volume is kept constant by adding a flow rate filter.
            The default is True.

        Raises
        ------
        CADETProcessError
            If compartment is not a real compartment.
        """
        tracer = Inlet(self.component_system, "tracer")
        tracer.flow_rate = flow_rate

        compartment_unit = self.flow_sheet[f"compartment_{compartment_index}"]
        if compartment_unit not in self._real_compartments:
            raise CADETProcessError("Tracer must connect to real compartment")

        if flow_rate_filter:
            compartment_unit.flow_rate_filter = flow_rate

        self.flow_sheet.add_unit(tracer)
        self.flow_sheet.add_connection(tracer, compartment_unit)

        self.process.add_event("tracer_on", "flow_sheet.tracer.c", c, t_start)
        self.process.add_event("tracer_off", "flow_sheet.tracer.c", 0, t_start + t_inj)

    def validate_flow_rates(self) -> None:
        """Validate that compartment volume is constant."""
        flow_rates = self.flow_sheet.get_flow_rates()

        for comp in self._real_compartments:
            for port in flow_rates[comp.name].total_in:
                if not np.all(
                    np.isclose(
                        flow_rates[comp.name].total_in[port],
                        flow_rates[comp.name].total_out[port],
                    )
                ):
                    if comp.n_ports == 1:
                        msg = comp.name
                    else:
                        msg = f"{comp.name} at Port {port}"
                    raise CADETProcessError(
                        f"Unbalanced flow rate for compartment '{msg}'."
                    )


class CompartmentModel(Cstr):
    """Dummy Class for checking binding and reaction models."""

    pass
